from .alpha import Alpha
from .anglit import Anglit
from .beta import Beta, ArcSinus, PowerLaw, Wigner, PERT
from .binomial import Binomial
from .bradford import Bradford
from .burr import Burr
from .cauchy import Cauchy
from .chi import Chi, Maxwell, Rayleigh
from .chi_squared import ChiSquared
from .discrete_uniform import DiscreteUniform
from .double_gamma import DoubleGamma
from .double_weibull import DoubleWeibull
from .exponential_power import ExponentialPower
from .exponential_weibull import ExponentialWeibull
from .f import F
from .fatigue_life import FatigueLife
from .fisk import Fisk
from .folded_cauchy import FoldedCauchy
from .folded_normal import FoldedNormal
from .frechet import Frechet
from .gamma import Gamma, Exponential
from .generalized_exponential import GeneralizedExponential
from .generalized_extreme import GeneralizedExtreme
from .generalized_gamma import GeneralizedGamma
from .generalized_half_logistic import GeneralizedHalfLogistic
from .gompertz import Gompertz
from .hyperbolic_secant import HyperbolicSecant
from .inverse_gamma import InverseGamma
from .kumaraswamy import Kumaraswamy
from .laplace import Laplace
from .levy import Levy
from .log_gamma import LogGamma
from .log_laplace import LogLaplace
from .log_normal import LogNormal, Gilbrat
from .log_uniform import LogUniform
from .log_weibull import LogWeibull
from .logistic import Logistic
from .mielke import Mielke
from .mv_log_normal import MvLogNormal
from .mv_normal import MvNormal
from .mv_student_t import MvStudentT
from .nakagami import Nakagami
from .normal import Normal
from .pareto1 import Pareto1
from .pareto2 import Pareto2
from .power_log_normal import PowerLogNormal
from .power_normal import PowerNormal
from .reciprocal import Reciprocal
from .student_t import StudentT
from .triangle import Triangle
from .trunc_exponential import TruncExponential
from .trunc_normal import TruncNormal
from .tukey_lambda import TukeyLambda
from .uniform import Uniform
from .wald import Wald
from .weibull import Weibull
from .wrapped_cauchy import WrappedCauchy



"""
Create samples from a Korobov lattice.

Examples usage
--------------

Normal usage::

    >>> distribution = chaospy.J(chaospy.Uniform(0, 1), chaospy.Uniform(0, 1))
    >>> samples = distribution.sample(4, rule="korobov")
    >>> samples.round(4)
    array([[0.2, 0.4, 0.6, 0.8],
           [0.4, 0.8, 0.2, 0.6]])
    >>> samples = distribution.sample(6, rule="korobov")
    >>> samples.round(4)
    array([[0.1429, 0.2857, 0.4286, 0.5714, 0.7143, 0.8571],
           [0.4286, 0.8571, 0.2857, 0.7143, 0.1429, 0.5714]])

"""
import numpy


def create_korobov_samples(order, dim, base=17797):
    """
    Create Korobov lattice samples.

    Args:
        order (int):
            The order of the Korobov latice. Defines the number of
            samples.
        dim (int):
            The number of dimensions in the output.
        base (int):
            The number based used to calculate the distribution of values.

    Returns (numpy.ndarray):
        Korobov lattice with ``shape == (dim, order)``
    """
    values = numpy.empty(dim)
    values[0] = 1
    for idx in range(1, dim):
        values[idx] = base*values[idx-1] % (order+1)

    grid = numpy.mgrid[:dim, :order+1]
    out = values[grid[0]] * (grid[1]+1) / (order+1.) % 1.
    return out[:, :order]



"""Variance operator."""
import numpy
import numpoly

from .expected import E


def Var(poly, dist=None, **kws):
    """
    Element by element 2nd order statistics.

    Args:
        poly (numpoly.ndpoly, Distribution):
            Input to take variance on.
        dist (Distribution):
            Defines the space the variance is taken on. It is ignored if
            ``poly`` is a distribution.

    Returns:
        (numpy.ndarray):
            Element for element variance along ``poly``, where
            ``variation.shape == poly.shape``.

    Examples:
        >>> dist = chaospy.J(chaospy.Gamma(1, 1), chaospy.Normal(0, 2))
        >>> chaospy.Var(dist)
        array([1., 4.])
        >>> q0, q1 = chaospy.variable(2)
        >>> poly = chaospy.polynomial([1, q0, q1, 10*q0*q1])
        >>> chaospy.Var(poly, dist)
        array([  0.,   1.,   4., 800.])
    """
    if dist is None:
        dist, poly = poly, numpoly.variable(len(poly))
    poly = numpoly.set_dimensions(poly, len(dist))
    if not poly.isconstant:

        return poly.tonumpy()**2
    poly = poly-E(poly, dist, **kws)
    poly = numpoly.square(poly)
    return E(poly, dist, **kws)



"""Gompertz distribution."""
import numpy
from scipy import special

from ..baseclass import SimpleDistribution, ShiftScaleDistribution


class gompertz(SimpleDistribution):
    """Gompertz distribution."""

    def __init__(self, c):
        super(gompertz, self).__init__(dict(c=c))

    def _pdf(self, x, c):
        ex = numpy.exp(x)
        return c*ex*numpy.exp(-c*(ex-1))

    def _cdf(self, x, c):
        return 1.0-numpy.exp(-c*(numpy.exp(x)-1))

    def _ppf(self, q, c):
        return numpy.log(1-1.0/c*numpy.log(1-q))

    def _lower(self, c):
        return 0.

    def _upper(self, c):
        return numpy.log(1+27.7/c)


class Gompertz(ShiftScaleDistribution):
    """
    Gompertz distribution

    Args:
        shape (float, Distribution):
            Shape parameter
        scale (float, Distribution):
            Scaling parameter
        shift (float, Distribution):
            Location parameter

    Examples:
        >>> distribution = chaospy.Gompertz(1.5)
        >>> distribution
        Gompertz(1.5)
        >>> uloc = numpy.linspace(0, 1, 6)
        >>> uloc
        array([0. , 0.2, 0.4, 0.6, 0.8, 1. ])
        >>> xloc = distribution.inv(uloc)
        >>> xloc.round(3)
        array([0.   , 0.139, 0.293, 0.477, 0.729, 2.969])
        >>> numpy.allclose(distribution.fwd(xloc), uloc)
        True
        >>> distribution.pdf(xloc).round(3)
        array([1.5  , 1.379, 1.206, 0.967, 0.622, 0.   ])
        >>> distribution.sample(4).round(3)
        array([0.535, 0.078, 1.099, 0.364])

    """

    def __init__(self, shape, scale=1, shift=0):
        super(Gompertz, self).__init__(
            dist=gompertz(shape),
            scale=scale,
            shift=shift,
            repr_args=[shape],
        )



"""Beta distribution."""
import numpy
from scipy import special
import chaospy

from ..baseclass import SimpleDistribution, LowerUpperDistribution, ShiftScaleDistribution


class beta_(SimpleDistribution):

    def __init__(self, a=1, b=1):
        super(beta_, self).__init__(dict(a=a, b=b))

    def _pdf(self, x, a, b):
        out = x**(a-1)*(1-x)**(b-1)/special.beta(a, b)
        out = numpy.where(numpy.isfinite(out), out, 0)
        return out

    def _cdf(self, x, a, b):
        return special.btdtr(a, b, x)

    def _ppf(self, qloc, a, b):
        return special.btdtri(a, b, qloc)

    def _mom(self, k, a, b):
        return special.beta(a+k,b)/special.beta(a,b)

    def _ttr(self, n, a, b):
        nab = 2*n+a+b
        A = ((a-1)**2-(b-1)**2)*.5/\
                (nab*(nab-2) + (nab==0) + (nab==2)) + .5
        B1 = a*b*1./((a+b+1)*(a+b)**2)
        B2 = (n+a-1)*(n+b-1)*n*(n+a+b-2.)/\
            ((nab-1)*(nab-3)*(nab-2)**2+2.*((n==0)+(n==1)))
        B = numpy.where((n==0)+(n==1), B1, B2)
        return A, B

    def _lower(self, a, b):
        return 0.

    def _upper(self, a, b):
        return 1.


class Beta(LowerUpperDistribution):
    R"""
    Beta Probability Distribution.

    Args:
        alpha (float, Distribution):
            First shape parameter, alpha > 0
        beta (float, Distribution):
            Second shape parameter, b > 0
        lower (float, Distribution):
            Lower threshold
        upper (float, Distribution):
            Upper threshold

    Examples:
        >>> distribution = chaospy.Beta(1.5, 3.5)
        >>> distribution
        Beta(1.5, 3.5)
        >>> uloc = numpy.linspace(0, 1, 6)
        >>> uloc
        array([0. , 0.2, 0.4, 0.6, 0.8, 1. ])
        >>> xloc = distribution.inv(uloc)
        >>> xloc.round(3)
        array([0.   , 0.126, 0.222, 0.326, 0.464, 1.   ])
        >>> numpy.allclose(distribution.fwd(xloc), uloc)
        True
        >>> distribution.pdf(xloc).round(3)
        array([0.   , 2.066, 2.051, 1.734, 1.168, 0.   ])
        >>> distribution.sample(4).round(3)
        array([0.358, 0.083, 0.651, 0.263])
        >>> distribution.mom(1).round(4)
        0.3
        >>> distribution.ttr([0, 1, 2, 3]).round(4)
        array([[0.3   , 0.4143, 0.4524, 0.4697],
               [0.035 , 0.035 , 0.0478, 0.0535]])

    """

    def __init__(self, alpha, beta, lower=0, upper=1):
        super(Beta, self).__init__(
            dist=beta_(alpha, beta),
            lower=lower,
            upper=upper,
            repr_args=[alpha, beta],
        )


class ArcSinus(LowerUpperDistribution):
    """
    Generalized Arc-sinus distribution

    Args:
        shape (float, Distribution):
            Shape parameter where 0.5 is the default non-generalized case.
            Defined on the interval ``[0, 1]``.
        lower (float, Distribution):
            Lower threshold
        upper (float, Distribution):
            Upper threshold

    Examples:
        >>> distribution = chaospy.ArcSinus(0.5)
        >>> distribution
        ArcSinus(0.5)
        >>> uloc = numpy.linspace(0, 1, 6)
        >>> uloc
        array([0. , 0.2, 0.4, 0.6, 0.8, 1. ])
        >>> xloc = distribution.inv(uloc)
        >>> xloc.round(3)
        array([0.   , 0.095, 0.345, 0.655, 0.905, 1.   ])
        >>> numpy.allclose(distribution.fwd(xloc), uloc)
        True
        >>> distribution.pdf(xloc).round(3)
        array([0.   , 1.083, 0.669, 0.669, 1.083, 0.   ])
        >>> distribution.sample(4).round(3)
        array([0.732, 0.032, 0.994, 0.472])
        >>> distribution.mom(1).round(4)
        0.5
        >>> distribution.ttr([0, 1, 2, 3]).round(4)
        array([[0.5   , 0.5   , 0.5   , 0.5   ],
               [0.125 , 0.125 , 0.0625, 0.0625]])

    """

    def __init__(self, shape=0.5, lower=0, upper=1):
        super(ArcSinus, self).__init__(
            dist=beta_(shape, 1-shape),
            lower=lower,
            upper=upper,
            repr_args=[shape],
        )


class PowerLaw(LowerUpperDistribution):
    """
    Powerlaw distribution

    Args:
        shape (float, Distribution):
            Shape parameter
        lower (float, Distribution):
            Location of lower threshold
        upper (float, Distribution):
            Location of upper threshold

    Examples:
        >>> distribution = chaospy.PowerLaw(0.8)
        >>> distribution
        PowerLaw(0.8)
        >>> uloc = numpy.linspace(0, 1, 6)
        >>> uloc
        array([0. , 0.2, 0.4, 0.6, 0.8, 1. ])
        >>> xloc = distribution.inv(uloc)
        >>> xloc.round(3)
        array([0.   , 0.134, 0.318, 0.528, 0.757, 1.   ])
        >>> numpy.allclose(distribution.fwd(xloc), uloc)
        True
        >>> distribution.pdf(xloc).round(3)
        array([0.   , 1.196, 1.006, 0.909, 0.846, 0.8  ])
        >>> distribution.sample(4).round(3)
        array([0.588, 0.067, 0.938, 0.402])
        >>> distribution.mom(1).round(4)
        0.4444
        >>> distribution.ttr([0, 1, 2, 3]).round(4)
        array([[0.4444, 0.5029, 0.5009, 0.5004],
               [0.0882, 0.0882, 0.0668, 0.0643]])

    """

    def __init__(self, shape=1, lower=0, upper=1):
        super(PowerLaw, self).__init__(
            dist=beta_(shape, 1),
            lower=lower,
            upper=upper,
            repr_args=[shape],
        )


class Wigner(ShiftScaleDistribution):
    """
    Wigner (semi-circle) distribution

    Args:
        radius (float, Distribution):
            Radius of the semi-circle (scale)
        shift (float, Distribution):
            Location of the circle origin (location)

    Examples:
        >>> distribution = chaospy.Wigner(1.5)
        >>> distribution
        Wigner(1.5)
        >>> uloc = numpy.linspace(0, 1, 6)
        >>> uloc
        array([0. , 0.2, 0.4, 0.6, 0.8, 1. ])
        >>> xloc = distribution.inv(uloc)
        >>> xloc.round(3)
        array([-1.5  , -0.738, -0.237,  0.237,  0.738,  1.5  ])
        >>> numpy.allclose(distribution.fwd(xloc), uloc)
        True
        >>> distribution.pdf(xloc).round(3)
        array([0.   , 0.37 , 0.419, 0.419, 0.37 , 0.   ])
        >>> distribution.sample(4).round(3)
        array([ 0.366, -0.983,  1.209, -0.042])
        >>> distribution.mom(1).round(4)
        0.0
        >>> distribution.ttr([0, 1, 2, 3]).round(4)
        array([[0.    , 0.    , 0.    , 0.    ],
               [0.5625, 0.5625, 0.5625, 0.5625]])

    """

    def __init__(self, radius=1, shift=0):
        super(Wigner, self).__init__(
            dist=beta_(1.5, 1.5),
            scale=2*radius, shift=shift-radius)
        self._repr_args = [radius]+chaospy.format_repr_kwargs(shift=(shift, 0))


class PERT(Beta):
    r"""
    Program Evaluation and Review Technique (PERT) Distribution.

    Defined by its mean::

        \mu = \frac{lower + gamma*mode + upper}{2 + gamma}

    Normal PERT for `gamma=4`. Other values results in the so called
    modified-PERT distribution.

    Args:
        lower (float):
            The lower bounds for the distribution.
        mode (float, Distribution):
            The mode of the distribution.
        upper (float):
            The upper bounds for the distribution.
        gamma (flat, Distribution):
            Modify the PERT distribution to make more emphasis on the
            distribution mode instead of the distribution tails.

    Examples:
        >>> distribution = chaospy.PERT(-1, 0, 1)
        >>> distribution
        PERT(-1, 0, 1)
        >>> uloc = numpy.linspace(0, 1, 6)
        >>> uloc
        array([0. , 0.2, 0.4, 0.6, 0.8, 1. ])
        >>> xloc = distribution.inv(uloc)
        >>> xloc.round(3)
        array([-1.   , -0.347, -0.107,  0.107,  0.347,  1.   ])
        >>> numpy.allclose(distribution.fwd(xloc), uloc)
        True
        >>> distribution.pdf(xloc).round(3)
        array([0.   , 0.726, 0.916, 0.916, 0.726, 0.   ])
        >>> distribution.sample(4).round(3)
        array([ 0.167, -0.479,  0.622, -0.019])
        >>> distribution.mom(1).round(4)
        0.0
        >>> distribution.ttr([0, 1, 2, 3]).round(4)
        array([[0.    , 0.    , 0.    , 0.    ],
               [0.1429, 0.1429, 0.1905, 0.2121]])

    """

    def __init__(self, lower, mode, upper, gamma=4):
        mu = (lower+4*mode+upper)/6.
        alpha = 1+gamma*(mu-lower)/(upper-lower)
        beta = 1+gamma*(upper-mu)/(upper-lower)
        LowerUpperDistribution.__init__(
            self,
            dist=beta_(alpha, beta),
            lower=lower,
            upper=upper,
        )
        self._repr_args = [lower, mode, upper]
        self._repr_args += chaospy.format_repr_kwargs(gamma=(gamma, 4))



"""Log-uniform distribution."""
import numpy

from ..baseclass import SimpleDistribution, ShiftScaleDistribution


class log_uniform(SimpleDistribution):
    """Log-uniform distribution."""

    def __init__(self, lo=0, up=1):
        super(log_uniform, self).__init__(dict(lo=lo, up=up))

    def _pdf(self, x, lo, up):
        return 1./(x*(up-lo))

    def _cdf(self, x, lo, up):
        return (numpy.log(x)-lo)/(up-lo)

    def _ppf(self, q, lo, up):
        return numpy.e**(q*(up-lo) + lo)

    def _lower(self, lo, up):
        return numpy.e**lo

    def _upper(self, lo, up):
        return numpy.e**up

    def _mom(self, k, lo, up):
        return ((numpy.e**(up*k)-numpy.e**(lo*k))/((up-lo)*(k+(k==0))))**(k!=0)


class LogUniform(ShiftScaleDistribution):
    """
    Log-uniform distribution

    Args:
        lower (float, Distribution):
            Location of lower threshold of uniform distribution.
        upper (float, Distribution):
            Location of upper threshold of uniform distribution.
        scale (float, Distribution):
            Scaling parameter
        shift (float, Distribution):
            Location parameter

    Examples:
        >>> distribution = chaospy.LogUniform(0.5, 1.5)
        >>> distribution
        LogUniform(0.5, 1.5)
        >>> uloc = numpy.linspace(0, 1, 6)
        >>> uloc
        array([0. , 0.2, 0.4, 0.6, 0.8, 1. ])
        >>> xloc = distribution.inv(uloc)
        >>> xloc.round(3)
        array([1.649, 2.014, 2.46 , 3.004, 3.669, 4.482])
        >>> numpy.allclose(distribution.fwd(xloc), uloc)
        True
        >>> distribution.pdf(xloc).round(3)
        array([0.607, 0.497, 0.407, 0.333, 0.273, 0.223])
        >>> distribution.sample(4).round(3)
        array([3.17 , 1.85 , 4.264, 2.67 ])
        >>> distribution.mom(1).round(3)
        2.833

    """

    def __init__(self, lower=0, upper=1, scale=1, shift=0):
        super(LogUniform, self).__init__(
            dist=log_uniform(lower, upper),
            scale=scale,
            shift=shift,
            repr_args=[lower, upper],
        )



"""Frechet or Extreme value distribution type 2."""
import numpy
from scipy import special

from ..baseclass import SimpleDistribution, ShiftScaleDistribution


class frechet(SimpleDistribution):
    """Frechet or Extreme value distribution type 2."""

    def __init__(self, c=1):
        super(frechet, self).__init__(dict(c=c))

    def _pdf(self, x, c):
        return c*pow(x,c-1)*numpy.exp(-pow(x,c))

    def _cdf(self, x, c):
        return -numpy.expm1(-pow(x,c))

    def _ppf(self, q, c):
        return pow(-numpy.log1p(-q), 1./c)

    def _mom(self, k, c):
        return special.gamma(1-k*1./c)

    def _lower(self, c):
        return 0.

    def _upper(self, c):
        return pow(35, (1./c))


class Frechet(ShiftScaleDistribution):
    """
    Frechet or Extreme value distribution type 2.

    Args:
        shape (float, Distribution):
            Shape parameter.
        scale (float, Distribution):
            Scaling parameter.
        shift (float, Distribution):
            Location parameter.

    Examples:
        >>> distribution = chaospy.Frechet(3)
        >>> distribution
        Frechet(3)
        >>> uloc = numpy.linspace(0, 1, 6)
        >>> uloc
        array([0. , 0.2, 0.4, 0.6, 0.8, 1. ])
        >>> xloc = distribution.inv(uloc)
        >>> xloc.round(3)
        array([0.   , 0.607, 0.799, 0.971, 1.172, 3.271])
        >>> numpy.allclose(distribution.fwd(xloc), uloc)
        True
        >>> distribution.pdf(xloc).round(3)
        array([0.   , 0.883, 1.15 , 1.132, 0.824, 0.   ])
        >>> distribution.sample(4).round(3)
        array([1.02 , 0.496, 1.442, 0.87 ])

    """

    def __init__(self, shape=1, scale=1, shift=0):
        super(Frechet, self).__init__(
            dist=frechet(shape),
            scale=scale,
            shift=shift,
            repr_args=[shape],
        )



"""Example code used in the tutorials."""
import numpy
import chaospy

coordinates = numpy.linspace(0, 10, 1000)

distribution_init = chaospy.Normal(1.5, 0.2)
distribution_rate = chaospy.Uniform(0.1, 0.2)
distribution = chaospy.J(distribution_init, distribution_rate)

_t = coordinates[1:]
true_mean = numpy.hstack([1.5, 15*(numpy.e**(-0.1*_t)-numpy.e**(-0.2*_t))/_t])
true_variance = numpy.hstack([2.29, 11.45*(numpy.e**(-0.2*_t)-numpy.e**(-0.4*_t))/_t])-true_mean**2


def exponential_model(parameters, coordinates=coordinates):
    """
    Over simplistic exponential model function.

    Args:
        parameters (numpy.ndarray):
            Hyper-parameters defining the model initial conditions and
            exponential growth rate.
            Assumed to have ``parameters.shape == (2,)``.
        coordinates (numpy.ndarray):
            The spatio-temporal coordinates.

    Returns:
        (numpy.ndarray):
            Evaluation of the exponential model. Same shape as ``coordinates``.
    """
    param_init, param_rate = parameters
    return param_init*numpy.e**(-param_rate*coordinates)


def error_mean(prediction_mean, true_mean=true_mean):
    """
    How close the estimated mean is the to the true mean.

    Args:
        prediction_mean (numpy.ndarray):
            The estimated mean.
        true_mean (numpy.ndarray):
            The reference mean value. Must be same shape as
            ``prediction_mean``.

    Returns:
        (float):
            The mean absolute distance between predicted and true values.
    """
    return numpy.mean(numpy.abs(prediction_mean-true_mean))


def error_variance(predicted_variance, true_variance=true_variance):
    """
    How close the estimated variance is the to the true variance.

    Args:
        prediction_variance (numpy.ndarray):
            The estimated variance.
        true_variance (numpy.ndarray):
            The reference variance value. Must be same shape as
            ``prediction_variance``.

    Returns:
        (float):
            The mean absolute distance between predicted and true values.
    """
    return numpy.mean(numpy.abs(predicted_variance-true_variance))



"""
Generate the quadrature abscissas and weights for simple grid.

Mostly available to ensure that discrete distributions works along side
continuous ones.
"""
import numpy
import chaospy

from .combine import combine_quadrature


def quad_grid(order, domain=(0, 1)):
    """
    Generate the quadrature abscissas and weights for simple grid.

    Args:
        order (int, numpy.ndarray):
            Quadrature order.
        domain (chaospy.distributions.baseclass.Distribution, numpy.ndarray):
            Either distribution or bounding of interval to integrate over.

    Returns:
        (numpy.ndarray, numpy.ndarray):
            The quadrature points and weights. The points are
            equi-spaced grid on the interior of the domain bounds.
            The weights are all equal to `1/len(weights[0])`.

    Example:
        >>> abscissas, weights = chaospy.quad_grid(4, chaospy.Uniform(-1, 1))
        >>> abscissas.round(4)
        array([[-0.8, -0.4,  0. ,  0.4,  0.8]])
        >>> weights.round(4)
        array([0.2, 0.2, 0.2, 0.2, 0.2])
        >>> abscissas, weights = chaospy.quad_grid([1, 1])
        >>> abscissas.round(4)
        array([[0.25, 0.25, 0.75, 0.75],
               [0.25, 0.75, 0.25, 0.75]])
        >>> weights.round(4)
        array([0.25, 0.25, 0.25, 0.25])

    """
    if isinstance(domain, chaospy.Distribution):
        abscissas, weights = quad_grid(order, (domain.lower, domain.upper))
        weights *= domain.pdf(abscissas).flatten()
        weights /= numpy.sum(weights)
        return abscissas, weights

    order = numpy.atleast_1d(order)
    order, lower, upper = numpy.broadcast_arrays(order, domain[0], domain[1])
    assert order.ndim == 1, "too many dimensions"
    abscissas = tuple(numpy.linspace(0, 1, 2*order_+3)[1::2] for order_ in order)
    weights = tuple(numpy.repeat(1./(order_+1), order_+1) for order_ in order)

    abscissas_, weights_ = combine_quadrature(abscissas, weights, (lower, upper))
    return abscissas_, weights_



"""Gaussian kernel density estimation."""
import numpy
from scipy import special

from .baseclass import KernelDensityBaseclass
from ..collection.mv_normal import MvNormal


class GaussianKDE(KernelDensityBaseclass):
    """
    Gaussian kernel density estimator.

    Density estimator that handles both univariate and multivariate data. It
    provides automatic bandwidth selection method using Scott's and Silverman's
    method.

    Attributes:
        samples:
            The raw data as provided by the user reshaped to have `ndim == 2`.
        h_mat:
            The covariance matrix to each sample. Assuming uncorrelated
            dimensions, the bandwidth is the square root of the diagonals. It
            will have either dimensions `(1, n_dim, n_dim)` if all samples
            shares covariance, or `(n_samples, n_dim, n_dim)` if not.
        weights:
            How much each sample is weighted. Either a scalar when the samples
            are equally weighted, or with length `n_samples` otherwise.

    Examples:
        >>> samples = [[-1, 0, 1], [0, 1, 2]]
        >>> distribution = chaospy.GaussianKDE(samples, estimator_rule="silverman")
        >>> distribution.h_mat  # H-matrix or bandwidth**2
        array([[[0.38614462, 0.        ],
                [0.        , 0.38614462]]])
        >>> uloc = [[0, 0, 1, 1], [0, 1, 0, 1]]
        >>> distribution.pdf(uloc).round(4)
        array([0.0469, 0.0982, 0.0074, 0.0469])
        >>> distribution.fwd(uloc).round(4)
        array([[0.5   , 0.5   , 0.8152, 0.8152],
               [0.1233, 0.5   , 0.0142, 0.1532]])
        >>> distribution.inv(uloc).round(4)
        array([[-6.577 , -6.577 ,  5.3948,  5.3948],
               [-4.3871,  4.6411, -5.9611,  7.7779]])
        >>> distribution.mom([(0, 1, 1), (1, 0, 1)]).round(4)
        array([1.    , 0.    , 0.6667])

    """

    @staticmethod
    def _kernel(z_loc):
        """The underlying density kernel."""
        return numpy.prod(numpy.e**(-z_loc**2/2.)/numpy.sqrt(2*numpy.pi), axis=-1)

    @staticmethod
    def _ikernel(z_loc):
        """The integrand of the underlying density kernel."""
        return special.ndtr(z_loc)

    def _mom(self, k_loc, cache):
        """Raw statistical moments."""
        length = self.samples.shape[-1]
        h_mat = numpy.broadcast_to(
            self.h_mat, (length,)+self.h_mat.shape[1:])
        out = numpy.array([
            MvNormal._mom(
                self,
                k_loc,
                mean=self.samples[:, idx],
                sigma=h_mat[idx],
                cache={},
            )
            for idx in range(length)
        ])
        return numpy.sum(out*self.weights)

    def _lower(self, idx, dim, cache):
        """Lower bounds."""
        return (self.samples[idx]-10*numpy.sqrt(self.h_mat[:, idx, idx]).T).min(-1)

    def _upper(self, idx, dim, cache):
        """Upper bounds."""
        return (self.samples[idx]+10*numpy.sqrt(self.h_mat[:, idx, idx]).T).max(-1)



"""Cauchy distribution."""
import numpy

from ..baseclass import SimpleDistribution, ShiftScaleDistribution


class cauchy(SimpleDistribution):
    """Standard Cauchy distribution."""

    def __init__(self):
        super(cauchy, self).__init__()

    def _pdf(self, x):
        return 1.0/numpy.pi/(1.0+x*x)

    def _cdf(self, x):
        return 0.5 + 1.0/numpy.pi*numpy.arctan(x)

    def _ppf(self, q):
        return numpy.tan(numpy.pi*q-numpy.pi/2.0)

    def _lower(self):
        return -3e13

    def _upper(self):
        return 3e13


class Cauchy(ShiftScaleDistribution):
    """
    Cauchy distribution.

    Also known as Lorentz distribution, Cachy-Lorentz distribution, and
    Breit-Wigner distribution.

    Args:
        shift (float, Distribution):
            Location parameter
        scale (float, Distribution):
            Scaling parameter

    Examples:
        >>> distribution = chaospy.Cauchy()
        >>> distribution
        Cauchy()
        >>> uloc = numpy.linspace(0.1, 0.9, 5)
        >>> uloc
        array([0.1, 0.3, 0.5, 0.7, 0.9])
        >>> xloc = distribution.inv(uloc)
        >>> xloc.round(3)
        array([-3.078, -0.727,  0.   ,  0.727,  3.078])
        >>> numpy.allclose(distribution.fwd(xloc), uloc)
        True
        >>> distribution.pdf(xloc).round(3)
        array([0.03 , 0.208, 0.318, 0.208, 0.03 ])
        >>> distribution.sample(4).round(3)
        array([ 0.524, -2.646,  6.35 , -0.056])

    Notes:
        The Cauchy distribution is what is known as a "pathological"
        distribution. It is not only infinitely bound, but heavy tailed
        enough that approximate bounds is also infinite for any reasonable
        approximation. This makes both bounds and moments results in
        non-sensibel results. E.g.::

            >>> distribution.lower < -1e10
            array([ True])
            >>> distribution.upper > 1e10
            array([ True])

    """

    def __init__(self, scale=1, shift=0):
        super(Cauchy, self).__init__(
            dist=cauchy(),
            scale=scale,
            shift=shift,
        )



import numpy
import numpoly

from ..conditional import E_cond
from ..expected import E
from ..variance import Var


def Sens_m2(poly, dist, **kws):
    """
    Variance-based decomposition/Sobol' indices.

    Second order sensitivity indices.

    Args:
        poly (numpoly.ndpoly):
            Polynomial to find second order Sobol indices on.
        dist (Distribution):
            The distributions of the input used in ``poly``.

    Returns:
        (numpy.ndarray):
            First order sensitivity indices for each parameters in ``poly``,
            with shape ``(len(dist), len(dist)) + poly.shape``.

    Examples:
        >>> q0, q1 = chaospy.variable(2)
        >>> poly = chaospy.polynomial([1, q0*q1, q0**3*q1, q0*q1**3])
        >>> dist = chaospy.Iid(chaospy.Uniform(0, 1), 2)
        >>> chaospy.Sens_m2(poly, dist).round(4)
        array([[[0.    , 0.    , 0.    , 0.    ],
                [0.    , 0.1429, 0.2093, 0.2093]],
        <BLANKLINE>
               [[0.    , 0.1429, 0.2093, 0.2093],
                [0.    , 0.    , 0.    , 0.    ]]])

    """
    dim = len(dist)
    poly = numpoly.set_dimensions(poly, len(dist))

    out = numpy.zeros((dim, dim) + poly.shape)
    variance = Var(poly, dist)

    valids = variance != 0
    if not numpy.all(valids):
        out[:, :, valids] = Sens_m2(poly[valids], dist, **kws)
        return out

    conditional_v = numpy.zeros((dim,)+poly.shape)
    for idx, unit_vec in enumerate(numpy.eye(dim, dtype=int)):
        conditional_e = E_cond(poly, unit_vec, dist, **kws)
        conditional_v[idx] = Var(conditional_e, dist, **kws)

    for idx, unit_vec1 in enumerate(numpy.eye(dim, dtype=int)):
        for idy, unit_vec2 in enumerate(numpy.eye(dim, dtype=int)[idx+1:], idx+1):

            conditional_e = E_cond(poly, unit_vec1+unit_vec2, dist, **kws)
            out[idx, idy] = Var(conditional_e, dist, **kws)
            out[idx, idy] -= conditional_v[idx]
            out[idx, idy] -= conditional_v[idy]
            out[idx, idy] /= variance

    # copy upper matrix triangle down to lower triangle
    indices = numpy.tril_indices(dim, -1)
    out[indices] = numpy.swapaxes(out, 0, 1)[indices]

    return out



"""Gumbel or Log-Weibull distribution."""
import numpy

from ..baseclass import SimpleDistribution, ShiftScaleDistribution


class log_weibull(SimpleDistribution):
    """Gumbel or Log-Weibull distribution."""

    def __init__(self):
        super(log_weibull, self).__init__()

    def _pdf(self, x):
        ex = numpy.exp(-x)
        return ex*numpy.exp(-ex)

    def _cdf(self, x):
        return numpy.exp(-numpy.exp(-x))

    def _ppf(self, q):
        return -numpy.log(-numpy.log(q))

    def _lower(self):
        return -3.5

    def _upper(self):
        return 35


class LogWeibull(ShiftScaleDistribution):
    """
    Gumbel or Log-Weibull distribution.

    Args:
        scale (float, Distribution):
            Scaling parameter
        loc (float, Distribution):
            Location parameter

    Examples:
        >>> distribution = chaospy.LogWeibull()
        >>> distribution
        LogWeibull()
        >>> uloc = numpy.linspace(0, 1, 6)
        >>> uloc
        array([0. , 0.2, 0.4, 0.6, 0.8, 1. ])
        >>> xloc = distribution.inv(uloc)
        >>> xloc.round(3)
        array([-3.5  , -0.476,  0.087,  0.672,  1.5  , 35.   ])
        >>> numpy.allclose(distribution.fwd(xloc), uloc)
        True
        >>> distribution.pdf(xloc).round(3)
        array([0.   , 0.322, 0.367, 0.306, 0.179, 0.   ])
        >>> distribution.sample(4).round(3)
        array([ 0.855, -0.771,  2.976,  0.316])

    """
    def __init__(self, scale=1, shift=0):
        super(LogWeibull, self).__init__(
            dist=log_weibull(),
            scale=scale,
            shift=shift,
        )



"""Generalized half-logistic distribution."""
import numpy
from scipy import special

from ..baseclass import SimpleDistribution, ShiftScaleDistribution


class generalized_half_logistic(SimpleDistribution):
    """Generalized half-logistic distribution."""

    def __init__(self, c=1):
        super(generalized_half_logistic, self).__init__(dict(c=c))

    def _pdf(self, x, c):
        limit = 1.0/c
        tmp = (1-c*x)
        tmp0 = tmp**(limit-1)
        tmp2 = tmp0*tmp
        return 2*tmp0/(1+tmp2)**2

    def _cdf(self, x, c):
        limit = 1.0/c
        tmp = (1-c*x)
        tmp2 = tmp**(limit)
        return (1.0-tmp2)/(1+tmp2)

    def _ppf(self, q, c):
        return 1.0/c*(1-((1.0-q)/(1.0+q))**c)

    def _lower(self, c):
        return 0.0

    def _upper(self, c):
        return 1/numpy.where(c < 10**-10, 10**-10, c)


class GeneralizedHalfLogistic(ShiftScaleDistribution):
    """
    Generalized half-logistic distribution

    Args:
        shape (float, Distribution):
            Shape parameter
        scale (float, Distribution):
            Scaling parameter
        shift (float, Distribution):
            Location parameter

    Examples:
        >>> distribution = chaospy.GeneralizedHalfLogistic(0.5)
        >>> distribution
        GeneralizedHalfLogistic(0.5)
        >>> uloc = numpy.linspace(0, 1, 6)
        >>> uloc
        array([0. , 0.2, 0.4, 0.6, 0.8, 1. ])
        >>> xloc = distribution.inv(uloc)
        >>> xloc.round(3)
        array([0.   , 0.367, 0.691, 1.   , 1.333, 2.   ])
        >>> numpy.allclose(distribution.fwd(xloc), uloc)
        True
        >>> distribution.pdf(xloc).round(3)
        array([0.5  , 0.588, 0.642, 0.64 , 0.54 , 0.   ])
        >>> distribution.sample(4).round(3)
        array([1.085, 0.218, 1.681, 0.818])

    """

    def __init__(self, shape, scale=1, shift=0):
        super(GeneralizedHalfLogistic, self).__init__(
            dist=generalized_half_logistic(shape),
            scale=scale,
            shift=shift,
            repr_args=[shape],
        )



"""Nataf (normal) copula."""
import numpy
from scipy import special
import chaospy

from ..baseclass import CopulaDistribution, Distribution


class nataf(Distribution):
    """Nataf (normal) copula."""

    def __init__(self, covariance, rotation=None):
        covariance = numpy.asarray(covariance)
        assert covariance.ndim == 2, "Covariance must be a matrix"
        assert covariance.shape[0] == covariance.shape[1], (
            "Parameters 'covariance' not a square matrix.")

        dependencies, _, rotation = chaospy.declare_dependencies(
            self,
            parameters=dict(covariance=covariance),
            rotation=rotation,
            dependency_type="accumulate",
        )
        correlation = covariance/numpy.sqrt(numpy.outer(numpy.diag(covariance), numpy.diag(covariance)))
        self._permute = numpy.eye(len(rotation), dtype=int)[rotation]
        self._correlation = self._permute.dot(correlation).dot(self._permute.T)
        cholesky = numpy.linalg.cholesky(self._correlation)
        self._fwd_transform = self._permute.T.dot(numpy.linalg.inv(cholesky))
        self._inv_transform = self._permute.T.dot(cholesky)

        super(nataf, self).__init__(
            parameters=dict(),
            dependencies=dependencies,
            rotation=rotation,
            repr_args=[covariance.tolist()],
        )

    def _cdf(self, xloc, idx, cache):
        dim = self._rotation.index(idx)
        conditions = [self._get_cache(dim_, cache, get=0)
                      for dim_ in self._rotation[:dim]]
        assert not any([isinstance(condition, chaospy.Distribution)
                        for condition in conditions])
        xloc = numpy.vstack(conditions+[xloc])
        zloc = self._fwd_transform[idx, :len(xloc)].dot(special.ndtri(xloc))
        out = special.ndtr(zloc)
        return out

    def _ppf(self, qloc, idx, cache):
        dim = self._rotation.index(idx)
        conditions = [self._get_cache(dim_, cache, get=1)
                      for dim_ in self._rotation[:dim]]
        assert not any([isinstance(condition, chaospy.Distribution)
                        for condition in conditions])
        qloc = numpy.vstack(conditions+[qloc])
        zloc = special.ndtri(qloc)
        out = special.ndtr(self._inv_transform[idx, :len(qloc)].dot(zloc))
        return out

    def _pdf(self, xloc, idx, cache):
        raise chaospy.UnsupportedFeature("Copula not supported.")

    def _lower(self, idx, cache):
        return 0.

    def _upper(self, idx, cache):
        return 1.


class Nataf(CopulaDistribution):
    """
    Nataf (normal) copula.

    Args:
        dist (Distribution):
            The distribution to wrap.
        covariance (numpy.ndarray):
            Covariance matrix.

    Examples:
        >>> distribution = chaospy.Nataf(
        ...     chaospy.Iid(chaospy.Uniform(-1, 1), 2), covariance=[[1, .5], [.5, 1]])
        >>> distribution
        Nataf(Iid(Uniform(lower=-1, upper=1), 2), [[1.0, 0.5], [0.5, 1.0]])
        >>> samples = distribution.sample(3)
        >>> samples.round(4)
        array([[ 0.3072, -0.77  ,  0.9006],
               [ 0.1262,  0.3001,  0.1053]])
        >>> distribution.pdf(samples).round(4)
        array([0.292 , 0.1627, 0.2117])
        >>> distribution.fwd(samples).round(4)
        array([[0.6536, 0.115 , 0.9503],
               [0.4822, 0.8725, 0.2123]])
        >>> mesh = numpy.meshgrid([.4, .5, .6], [.4, .5, .6])
        >>> distribution.inv(mesh).round(4)
        array([[[-0.2   ,  0.    ,  0.2   ],
                [-0.2   ,  0.    ,  0.2   ],
                [-0.2   ,  0.    ,  0.2   ]],
        <BLANKLINE>
               [[-0.2707, -0.1737, -0.0739],
                [-0.1008,  0.    ,  0.1008],
                [ 0.0739,  0.1737,  0.2707]]])

    """

    def __init__(self, dist, covariance):
        assert len(dist) == len(covariance)
        return super(Nataf, self).__init__(
            dist=dist,
            trans=nataf(covariance, dist._rotation),
            repr_args=[dist, numpy.array(covariance).tolist()],
        )



"""Constructor to create a range of polynomials where the exponent vary."""
import logging
import numpoly


def prange(N=1, dim=1):
    """
    Constructor to create a range of polynomials where the exponent vary.

    Args:
        N (int):
            Number of polynomials in the array.
        dim (int):
            The dimension the polynomial should span.

    Returns:
        (numpoly.ndpoly):
            A polynomial array of length N containing simple polynomials with
            increasing exponent.

    Examples:
        >>> chaospy.prange(4)
        polynomial([1, q0, q0**2, q0**3])
        >>> chaospy.prange(4, dim=3)
        polynomial([1, q2, q2**2, q2**3])
    """
    logger = logging.getLogger(__name__)
    logger.warning("chaospy.prange is deprecated; use chaospy.monomial instead")
    return numpoly.monomial(N, names=["q%d" % (dim-1)])



"""Reciprocal distribution."""
import numpy

from ..baseclass import SimpleDistribution, ShiftScaleDistribution


class reciprocal(SimpleDistribution):

    def __init__(self, lower=1, upper=2):
        super(reciprocal, self).__init__(dict(lower=lower, upper=upper))

    def _pdf(self, x, lower, upper):
        return 1./(x*numpy.log(upper/lower))

    def _cdf(self, x, lower, upper):
        return numpy.log(x/lower)/numpy.log(upper/lower)

    def _ppf(self, q, lower, upper):
        return numpy.e**(q*numpy.log(upper/lower)+numpy.log(lower))

    def _lower(self, lower, upper):
        return lower

    def _upper(self, lower, upper):
        return upper

    def _mom(self, kloc, lower, upper):
        return (upper**kloc-lower**kloc)/(kloc*numpy.log(upper/lower))


class Reciprocal(ShiftScaleDistribution):
    """
    Reciprocal distribution.

    Args:
        lower (float, Distribution):
            Lower threshold of distribution. Must be smaller than ``upper``.
        upper (float, Distribution):
            Upper threshold of distribution.
        scale (float, Distribution):
            Scaling parameter
        shift (float, Distribution):
            Location parameter


    Examples:
        >>> distribution = chaospy.Reciprocal(2, 4)
        >>> distribution
        Reciprocal(2, 4)
        >>> uloc = numpy.linspace(0, 1, 6)
        >>> uloc
        array([0. , 0.2, 0.4, 0.6, 0.8, 1. ])
        >>> xloc = distribution.inv(uloc)
        >>> xloc.round(3)
        array([2.   , 2.297, 2.639, 3.031, 3.482, 4.   ])
        >>> numpy.allclose(distribution.fwd(xloc), uloc)
        True
        >>> distribution.pdf(xloc).round(3)
        array([0.721, 0.628, 0.547, 0.476, 0.414, 0.361])
        >>> distribution.sample(4).round(3)
        array([3.146, 2.166, 3.865, 2.794])
        >>> distribution.mom(1).round(4)
        2.8854

    """

    def __init__(self, lower, upper, shift=0, scale=1):
        super(Reciprocal, self).__init__(
            dist=reciprocal(lower, upper),
            shift=shift, scale=scale,
            repr_args=[lower, upper],
        )



"""Function to combine two dataset together with a tensor product."""
import numpy


def combine(args):
    """
    All linear combination of a list of list.

    Args:
        args (numpy.ndarray):
            List of input arrays.  Components to take linear combination of
            with ``args[i].shape == (N[i], M[i])`` where ``N`` is to be taken
            linear combination of and ``M`` is constant.  ``M[i]`` is set to
            1 if missing.

    Returns:
        (numpy.array):
            Matrix of combinations with
            ``shape == (numpy.prod(N), numpy.sum(M))``.

    Examples:
        >>> A, B = [1,2], [[4,4],[5,6]]
        >>> chaospy.quadrature.combine([A, B])
        array([[1, 4, 4],
               [1, 5, 6],
               [2, 4, 4],
               [2, 5, 6]])
    """
    args = [numpy.asarray(arg).reshape(len(arg), -1) for arg in args]
    shapes = [arg.shape for arg in args]

    size = numpy.prod(shapes, 0)[0]*numpy.sum(shapes, 0)[1]
    if size > 10**9:
        raise MemoryError("Too large sets")

    out = args[0]
    for arg in args[1:]:
        out = numpy.hstack([
            numpy.tile(out, len(arg)).reshape(-1, out.shape[1]),
            numpy.tile(arg.T, len(out)).reshape(arg.shape[1], -1).T,
        ])
    return out


def combine_quadrature(
        abscissas,
        weights,
        domain=(),
):
    """
    Create all linear combinations of all abscissas and weights. If ``domain``
    is provided, also scale from assumed (0, 1) to said domain.

    Args:
        abscissas (List[numpy.ndarray]):
            List of abscissas to be combined.
        weights (List[numpy.ndarray]):
            List of weights to be combined.
        domain (Optional[Tuple[numpy.ndarray, numpy.ndarray]]):
            Domain to scale to.

    Returns:
        (Tuple[numpy.ndarray, numpy.ndarray]):
            Same as ``abscissas`` and ``weights``, but combined and flatten
            such that ``abscissas.shape == (dim, len(weights))``.
    """
    dim = len(abscissas)
    abscissas = combine(abscissas)
    weights = combine(weights)

    if domain:
        abscissas = (domain[1]-domain[0])*abscissas + domain[0]
        weights = weights*(domain[1]-domain[0])

    abscissas = abscissas.T.reshape(dim, -1)
    weights = numpy.prod(weights, -1)

    assert len(weights.shape) == 1
    assert abscissas.shape == (dim,) + weights.shape

    return abscissas, weights



import numpy
from scipy.special import comb
import numpoly
import chaospy

from .distribution import Distribution


class MeanCovarianceDistribution(Distribution):
    """
    Mean-Covariance transformation.

    Transforms any distribution with zero mean and unit variance into provided
    `mean` and `covariance`. The transformation works on other distributions as
    well, but in those cases `mean` and `covariance` are likely not correct.

    Args:
        dist (Distribution):
            The underlying distribution to be scaled.
        mean (float, Sequence[float], Distribution):
            Mean vector.
        covariance (float, Sequence[Sequence[float]], Distribution):
            Covariance matrix or variance vector if scale is a 1-d vector.
            If omitted, assumed to be 1.
        rotation (Sequence[int], Sequence[Sequence[bool]]):
            The order of which to resolve conditionals. Either as a sequence of
            column rotations, or as a permutation matrix.
            Defaults to `range(len(distribution))` which is the same as
            `p(x0), p(x1|x0), p(x2|x0,x1), ...`.

    """

    def __init__(
            self,
            dist,
            mean=0,
            covariance=1,
            rotation=None,
            repr_args=None,
    ):
        assert isinstance(dist, Distribution), "'dist' should be a distribution"
        mean = mean if isinstance(mean, Distribution) else numpy.atleast_1d(mean)
        assert not isinstance(covariance, Distribution)
        length = max(len(dist), len(mean), len(covariance))

        if not isinstance(mean, Distribution):
            assert mean.ndim == 1, "Parameter 'mean' have too many dimensions"
            assert len(mean) == length
        assert len(mean) in (1, length)

        exclusion = dist._exclusion.copy()
        if len(dist) == 1 and length > 1:
            dist = chaospy.Iid(dist, length)

        covariance = numpy.asarray(covariance)
        assert covariance.ndim <= 2, (
            "Covariance must either be scalar, vector or matrix")
        if covariance.ndim == 0:
            covariance = numpy.eye(length)*covariance
        elif covariance.ndim == 1:
            covariance = numpy.diag(covariance)
        assert covariance.shape == (length, length), (
            "Parameters 'mean' and 'covariance' have shape mismatch.")
        assert len(covariance) == length

        if rotation is not None:
            rotation = list(rotation)
        else:
            rotation = [key for key, _ in sorted(enumerate(dist._dependencies),
                                                 key=lambda x: len(x[1]))]

        accumulant = set()
        dependencies = [deps.copy() for deps in dist._dependencies]
        for idx in rotation:
            accumulant.update(dist._dependencies[idx])
            dependencies[idx] = accumulant.copy()

        if isinstance(mean, Distribution):
            if len(mean) == 1:
                for dependency in dependencies:
                    dependency.update(mean._dependencies[0])
            else:
                for dep1, dep2 in zip(dependencies, mean._dependencies):
                    dep1.update(dep2)

        self._permute = numpy.eye(len(rotation), dtype=int)[rotation]
        self._covariance = covariance
        self._pcovariance = self._permute.dot(covariance).dot(self._permute.T)
        cholesky = numpy.linalg.cholesky(self._pcovariance)
        self._fwd_transform = self._permute.T.dot(numpy.linalg.inv(cholesky))
        self._inv_transform = self._permute.T.dot(cholesky)
        self._dist = dist

        super(MeanCovarianceDistribution, self).__init__(
            parameters=dict(mean=mean, covariance=covariance),
            dependencies=dependencies,
            rotation=rotation,
            exclusion=exclusion,
            repr_args=repr_args,
        )

    def get_parameters(self, idx, cache, assert_numerical=True):
        parameters = super(MeanCovarianceDistribution, self).get_parameters(
            idx, cache, assert_numerical=assert_numerical)

        mean = parameters["mean"]
        if idx is None:
            return dict(mean=mean, sigma=self._covariance, cache=cache)

        if isinstance(mean, Distribution):
            mean = [mean._get_cache(dim, cache, get=0)
                    for dim in range(len(mean))]
            if any([isinstance(condition, chaospy.Distribution)
                    for condition in mean]):
                raise chaospy.StochasticallyDependentError(
                    "Dangling dependency: %s | %s" % (self, mean))
            mean = numpy.array(mean)
        mean = mean[self._rotation]

        dim = self._rotation.index(idx)
        if dim:
            covinv = numpy.linalg.inv(self._pcovariance[:dim, :dim])
            sigma = numpy.sqrt(
                self._pcovariance[dim, dim]-
                self._pcovariance[dim, :dim].dot(covinv).dot(self._pcovariance[:dim, dim])
            )
            mu_transform = self._pcovariance[dim, :dim].dot(covinv)
            assert numpy.isfinite(sigma), (dim, self._pcovariance)
        else:
            sigma = numpy.sqrt(self._pcovariance[0, 0])
            mu_transform = 0

        return dict(idx=idx, mean=mean, sigma=sigma, dim=dim, mut=mu_transform, cache=cache)

    def _lower(self, idx, mean, sigma, dim, mut, cache):
        out = mean[dim]+sigma*self._dist._get_lower(idx, cache)
        return out

    def _upper(self, idx, mean, sigma, dim, mut, cache):
        out = mean[dim]+sigma*self._dist._get_upper(idx, cache)
        return out

    def _pdf(self, xloc, idx, mean, sigma, dim, mut, cache):
        if dim:
            conditions = [self._get_cache(dim_, cache, get=0)
                          for dim_ in self._rotation[:dim]]
            assert not any([isinstance(condition, chaospy.Distribution)
                            for condition in conditions])
            mean = mean[dim]+mut.dot((numpy.vstack(conditions).T-mean[:dim]).T)
        else:
            mean = mean[dim]
        zloc = (xloc-mean)/sigma
        out = self._dist._get_pdf(zloc, idx, cache)/sigma
        return out

    def _ppf(self, uloc, idx, mean, sigma, dim, mut, cache):
        del sigma
        conditions = [self._get_cache(dim_, cache, get=1)
                      for dim_ in self._rotation[:dim]]
        assert not any([isinstance(condition, chaospy.Distribution)
                        for condition in conditions])
        uloc = numpy.vstack(conditions+[uloc])
        zloc = numpy.zeros(uloc.shape)
        for idx0, idx1 in enumerate(self._rotation[:len(uloc)]):
            zloc[idx0] = self._dist._get_inv(uloc[idx0], idx1, cache)
        xloc = self._inv_transform[idx, :len(uloc)].dot(zloc)+mean[dim]
        return xloc

    def _cdf(self, xloc, idx, mean, sigma, dim, mut, cache):
        del sigma
        conditions = [self[dim_]._get_cache(0, cache, get=0)
                      for dim_ in self._rotation[:dim]]
        assert not any([isinstance(condition, chaospy.Distribution)
                        for condition in conditions])
        xloc = numpy.vstack(conditions+[xloc])
        zloc = self._fwd_transform[idx, :len(xloc)].dot((xloc.T-mean[:len(xloc)]).T)
        uloc = self._dist._get_fwd(zloc, idx, cache)
        return uloc

    def _mom(self, kloc, mean, sigma, cache):
        poly = numpoly.variable(len(self))
        cholesky = numpy.linalg.cholesky(self._covariance)
        poly = numpoly.sum(cholesky*poly, axis=-1)+mean

        poly = numpoly.set_dimensions(numpoly.prod(poly**kloc), len(self))
        out = sum(self._dist.mom(key)*coeff
                  for key, coeff in zip(poly.exponents, poly.coefficients))
        return out

    def _ttr(self, kloc, idx, mean, sigma, dim, mut, cache):
        if dim > 1:
            raise chaospy.StochasticallyDependentError(
                "TTR require stochastically independent components.")
        coeff0, coeff1 = self._dist._get_ttr(kloc, idx)
        coeff0 = coeff0*sigma+mean[dim]
        coeff1 = coeff1*sigma*sigma
        return coeff0, coeff1



r"""
Gumbel Copula.

As a short hand:

.. math::

    s_n = \sigma(u, n, \theta)

The calculation of the derivative of the ``iphi`` function:

.. math::
  \begin{eqnarray}
    iphi(u)     &= e^{-u^{1/\theta}}
    iphi'(u)    &= e^{-u^{1/theta}}*(-1/theta*u^{1/theta-1})
                = iphi(u)*s_1 \\
    iphi''(u)   &= iphi'(u)*s_1+iphi(u)*s_2
                = iphi(u)*s_1^2+iphi(u)*s_2
                = iphi(u)*(s_1^2+s_2) \\
    iphi'''(u)  &= iphi'(u)*(s_1^2+s_2)+iphi(u)*(2*s_1*s_2+s_3)
                = iphi(u)*s_1*(s_1^2+s_2)+iphi(u)*(2*s_1*s_2+s_3)
                = iphi(u)*(s_1^3+3*s_1*s_2+s_3) \\
    iphi''''(u) &= iphi'(u)*(s_1^3+3*s_1*s_2+s_3) +
                   iphi(u)*(3*s_1^2*s_2+3*s_2^2+3*s_1*s_3+s_4)
                = iphi(u)*s_1*(s_1^3+3*s_1*s_2+s_3) +
                  iphi(u)*(3*s_1^2*s_2+3*s_2^2+3*s_1*s_3+s_4)
                = iphi(u)*(s_1^4 + 3*s_1^2*s_2 + s_1*s_3 +
                           3*s_1^2*s_2 + 3*s_2^2 + 3*s_1*s_3 + s_4)
                = iphi(u)*(s_1^4 + 6*s_1^2*s_2 +
                           s_1*s_3 + 3*s_2^2 + 3*s_1*s_3 + s_4)
  \end{eqnarray}

The same formula defined recursively:

.. math::
  \begin{eqnarray}
    iphi'(u)    &= iphi(u)*\sigma(u, 1, \theta) \\
    iphi''(u)   &= d/du ( iphi(u)*\sigma(u, 1, \theta) )
                = iphi'(u)*\sigma(u, 1, \theta) +
                  iphi(u)*\sigma(u, 2, \theta) \\
    iphi'''(u)  &= d/du (iphi'(u)*\sigma(u, 1, \theta) +
                          iphi(u)*\sigma(u, 2, \theta))
                = iphi''(u)*\sigma(u, 1, \theta) +
                  iphi'(u)*\sigma(u, 2, \theta) +
                  iphi'(u)*\sigma(u, 2, \theta) +
                  iphi(u)*\sigma(u, 3, \theta)
                = iphi''(u)*\sigma(u, 1, \theta) +
                  2*iphi'(u)*\sigma(u, 2, \theta) +
                  iphi(u)*\sigma(u, 3, \theta) \\
    iphi''''(u) &= d/du (iphi''(u)*\sigma(u, 1, \theta) +
                         2*iphi'(u)*\sigma(u, 2, \theta) +
                         iphi(u)*\sigma(u, 3, \theta))
                = iphi'''(u)*\sigma(u, 1, \theta) +
                  iphi''(u)*\sigma(u, 2, \theta) +
                  2*iphi''(u)*\sigma(u, 2, \theta) +
                  2*iphi'(u)*\sigma(u, 3, \theta) +
                  iphi'(u)*\sigma(u, 3, \theta) +
                  iphi(u)*\sigma(u, 4, \theta)
                = iphi'''(u)*\sigma(u, 1, \theta) +
                  3*iphi''(u)*\sigma(u, 2, \theta) +
                  3*iphi'(u)*\sigma(u, 3, \theta) +
                  iphi(u)*\sigma(u, 4, \theta) \\
                & \vdots \\
    iphi^{n}(u) &= sum_{i=1}^{n} comb(n, i-1) iphi^{n-i}(u) \sigma(u, i, \theta)
  \end{eqnarray}
"""
try:
    from functools import lru_cache
except ImportError:
    from functools32 import lru_cache
import numpy
from scipy import special

from ..baseclass import CopulaDistribution
from .archimedean import Archimedean


class gumbel(Archimedean):
    """Gumbel backend."""

    def _phi(self, t_loc, theta):
        return (-numpy.log(t_loc))**theta

    def _delta_phi(self, t_loc, theta):
        return -theta*(-numpy.log(t_loc))**(theta-1)/t_loc

    def _inverse_phi(self, u_loc, theta, order):
        @lru_cache(None)
        def iphi(n):
            if n:
                return sum(special.comb(n, i-1)*iphi(n-i)*sigma(i)
                           for i in range(1, n+1))
            return numpy.e**(-u_loc**(1/theta))
        @lru_cache(None)
        def sigma(n):
            return self._sigma(u_loc, theta, n)
        return iphi(order)


class Gumbel(CopulaDistribution):
    r"""
    Gumbel Copula.

    .. math::
        \phi(x;th) = \frac{x^{-th}-1}{th}
        \phi^{-1}(q;th) = (q*th + 1)^{-1/th}

    where `th` (or theta) is defined on the interval `[1,inf)`.

    Args:
        dist (Distribution):
            The distribution to wrap
        theta (float):
            Copula parameter

    Examples:
        >>> distribution = chaospy.Gumbel(
        ...     chaospy.Iid(chaospy.Uniform(-1, 1), 2), theta=2)
        >>> distribution
        Gumbel(Iid(Uniform(lower=-1, upper=1), 2), theta=2)
        >>> samples = distribution.sample(3)
        >>> samples.round(4)
        array([[ 0.3072, -0.77  ,  0.9006],
               [ 0.4709,  0.2101,  0.8696]])
        >>> distribution.pdf(samples).round(4)
        array([7.9977000e+00, 5.0700000e-02, 3.6733563e+03])
        >>> distribution.fwd(samples).round(4)
        array([[0.6536, 0.115 , 0.9503],
               [0.4822, 0.8725, 0.2123]])
        >>> mesh = numpy.meshgrid([.4, .5, .6], [.4, .5, .6])
        >>> distribution.inv(mesh).round(4)
        array([[[-0.2   ,  0.    ,  0.2   ],
                [-0.2   ,  0.    ,  0.2   ],
                [-0.2   ,  0.    ,  0.2   ]],
        <BLANKLINE>
               [[-0.0022,  0.1573,  0.3174],
                [ 0.109 ,  0.2591,  0.4062],
                [ 0.2181,  0.3564,  0.489 ]]])

    """

    def __init__(self, dist, theta):
        super(Gumbel, self).__init__(
            dist=dist,
            trans=gumbel(len(dist), theta),
            repr_args=[dist, "theta=%s" % theta],
        )



"""Gram-Schmidt process for generating orthogonal polynomials."""
import logging

import numpy
import numpoly
import chaospy


def orth_gs(order, dist, normed=False, graded=True, reverse=True,
            retall=False, cross_truncation=1., **kws):
    """
    Gram-Schmidt process for generating orthogonal polynomials.

    Args:
        order (int, numpoly.ndpoly):
            The upper polynomial order. Alternative a custom polynomial basis
            can be used.
        dist (Distribution):
            Weighting distribution(s) defining orthogonality.
        normed (bool):
            If True orthonormal polynomials will be used instead of monic.
        graded (bool):
            Graded sorting, meaning the indices are always sorted by the index
            sum. E.g. ``q0**2*q1**2*q2**2`` has an exponent sum of 6, and will
            therefore be consider larger than both ``q0**2*q1*q2``,
            ``q0*q1**2*q2`` and ``q0*q1*q2**2``, which all have exponent sum of
            5.
        reverse (bool):
            Reverse lexicographical sorting meaning that ``q0*q1**3`` is
            considered bigger than ``q0**3*q1``, instead of the opposite.
        retall (bool):
            If true return numerical stabilized norms as well. Roughly the same
            as ``cp.E(orth**2, dist)``.
        cross_truncation (float):
            Use hyperbolic cross truncation scheme to reduce the number of
            terms in expansion.

    Returns:
        (chapspy.poly.ndpoly):
            The orthogonal polynomial expansion.

    Examples:
        >>> distribution = chaospy.J(chaospy.Normal(), chaospy.Normal())
        >>> polynomials, norms = chaospy.orth_gs(2, distribution, retall=True)
        >>> polynomials.round(10)
        polynomial([1.0, q1, q0, q1**2-1.0, q0*q1, q0**2-1.0])
        >>> norms.round(10)
        array([1., 1., 1., 2., 1., 2.])
        >>> polynomials = chaospy.orth_gs(2, distribution, normed=True)
        >>> polynomials.round(3)
        polynomial([1.0, q1, q0, 0.707*q1**2-0.707, q0*q1, 0.707*q0**2-0.707])

    """
    logger = logging.getLogger(__name__)
    dim = len(dist)

    if isinstance(order, int):
        if order == 0:
            return numpoly.polynomial(1)
        basis = numpoly.monomial(
            0,
            order+1,
            names=numpoly.variable(2).names,
            graded=graded,
            reverse=reverse,
            cross_truncation=cross_truncation,
        )
    else:
        basis = order

    basis = list(basis)

    polynomials = [basis[0]]

    norms = [1.]
    for idx in range(1, len(basis)):

        # orthogonalize polynomial:
        for idy in range(idx):
            orth = chaospy.E(basis[idx]*polynomials[idy], dist, **kws)
            basis[idx] = basis[idx]-polynomials[idy]*orth/norms[idy]

        norms_ = chaospy.E(basis[idx]**2, dist, **kws)
        if norms_ <= 0:
            logger.warning("Warning: Polynomial cutoff at term %d", idx)
            break

        norms.append(1. if normed else norms_)
        basis[idx] = basis[idx]/numpy.sqrt(norms_) if normed else basis[idx]
        polynomials.append(basis[idx])

    polynomials = chaospy.polynomial(polynomials).flatten()
    if retall:
        norms = numpy.array(norms)
        return polynomials, norms
    return polynomials



import logging

import numpy
import chaospy


def approximate_inverse(
        distribution,
        idx,
        qloc,
        bounds=None,
        cache=None,
        parameters=None,
        xloc0=None,
        iterations=300,
        tolerance=1e-12,
):
    """
    Calculate the approximation of the inverse Rosenblatt transformation.

    Uses a hybrid Newton-Raphson and binary search method to converge to the
    inverse values. Includes forward Rosenblatt transformations, probability
    density function (its derivative), and if not provided, boundary function.

    Args:
        distribution (Distribution):
            Distribution to estimate inverse Rosenblatt on.
        idx (int):
            The dimension to take approximation along.
        qloc (numpy.ndarray):
            Input values. All values must be on unit interval ``(0, 1)`` and
            ``qloc.shape == (dim,size)`` where dim is the number of dimensions
            in distribution and size is the number of values to calculate
            simultaneously.
        bounds (Optional[Tuple[numpy.ndarray, numpy.ndarray]]):
            Assuming lower and upper bounds is not available, this provides
            outer bounds for lower and upper to use instead.
        cache (Optional[Dict[Distribution, numpy.ndarray]]):
            Memory cache for the location in the evaluation so far.
        parameters (Optional[Dict[str, Any]]):
            The parameters to use. If omitted, get the parameters from
            distribution.
        iterations (int):
            The maximum number of iterations allowed.
        tolerance (float):
            Tolerance criterion determining convergence.

    Returns:
        (numpy.ndarray):
            Approximation of inverse Rosenblatt transformation.

    Example:
        >>> distribution = chaospy.Normal(1000, 10)
        >>> qloc = numpy.array([0.1, 0.2, 0.9])
        >>> inverse = distribution.inv(qloc)
        >>> inverse.round(4)
        array([ 987.1845,  991.5838, 1012.8155])
        >>> distribution._ppf = None
        >>> numpy.allclose(
        ...     approximate_inverse(distribution, 0, qloc), inverse)
        True

    """
    logger = logging.getLogger(__name__)
    logger.debug("init approximate_inverse: %s", distribution)
    logger.debug("cache: %s", cache)

    # lots of initial values:
    if cache is None:
        cache = {}
    if bounds is None:
        xlower = distribution._get_lower(idx, cache.copy())
        xupper = distribution._get_upper(idx, cache.copy())
    else:
        xlower = numpy.broadcast_to(bounds[0], qloc.shape)
        xupper = numpy.broadcast_to(bounds[1], qloc.shape)

    xloc = xlower+qloc*(xlower+xupper) if xloc0 is None else xloc0
    uloc = numpy.zeros(qloc.shape)
    ulower = -qloc
    uupper = 1-qloc
    indices = numpy.ones(qloc.shape[-1], dtype=bool)

    if parameters is None:
        parameters = distribution.get_parameters(idx, cache, assert_numerical=True)
    else:
        assert not any([isinstance(value, chaospy.Distribution)
                        for value in parameters.values()])
        for name, param in parameters.items():
            if isinstance(distribution._parameters[name], chaospy.Distribution):
                cache[(idx, distribution._parameters[name])] = param

    for idx_ in range(2*iterations):

        # evaluate function:
        uloc = numpy.where(
            indices, distribution._cdf(xloc, **parameters)-qloc, uloc)

        # convergence criteria:
        indices &= numpy.abs(uloc) > tolerance
        if not numpy.any(indices):
            break

        # narrow down boundaries:
        ulower = numpy.where(indices & (uloc < 0), uloc, ulower)
        xlower = numpy.where(indices & (uloc < 0), xloc, xlower)
        uupper = numpy.where(indices & (uloc > 0), uloc, uupper)
        xupper = numpy.where(indices & (uloc > 0), xloc, xupper)

        # Newton increment every second iteration:
        xloc_ = numpy.inf
        if idx_ % 2 == 0:
            derivative = distribution._pdf(xloc, **parameters)
            derivative = numpy.where(derivative, derivative, 1)
            xloc_ = xloc-uloc/derivative

        # use binary search if Newton increment is outside bounds:
        weight = numpy.random.random()
        xloc_ = numpy.where((xloc_ < xupper) & (xloc_ > xlower),
                            xloc_, weight*xupper+(1-weight)*xlower)
        xloc = numpy.where(indices, xloc_, xloc)

    else:
        logger.warning(
            "Too many iterations required to estimate inverse.")
        logger.info("%d out of %d did not converge.",
            numpy.sum(indices), len(indices))
        # print("Too many iterations required to estimate inverse.")
        # print("%d out of %d did not converge." % (numpy.sum(indices), len(indices)))

    logger.debug("%s: ppf approx used %d steps", distribution, idx_/2)
    # print("%s: ppf approx used %d steps" % (distribution, idx_/2))
    return xloc

MOMENTS_QUADS = {}


def approximate_moment(
        distribution,
        k_loc,
        order=None,
        rule="fejer",
        **kwargs
):
    """
    Approximation method for estimation of raw statistical moments.

    Uses quadrature integration to estimate the values.

    Args:
        distribution (Distribution):
            Distribution domain with dim=len(distribution)
        k_loc (Sequence[int, ...]):
            The exponents of the moments of interest with ``shape == (dim,)``.
        order (int):
            The quadrature order used in approximation. If omitted, calculated
            to be ``1000/log2(len(distribution)+1)``.
        rule (str):
            Quadrature rule for integrating moments.
        kwargs:
            Extra args passed to `chaospy.generate_quadrature`.

    Examples:
        >>> distribution = chaospy.Uniform(1, 4)
        >>> round(chaospy.approximate_moment(distribution, (1,)), 4)
        2.5
        >>> round(chaospy.approximate_moment(distribution, (2,)), 4)
        7.0

    """
    if order is None:
        order = int(1000./numpy.log2(len(distribution)+1))
    assert isinstance(order, int)
    assert isinstance(distribution, chaospy.Distribution)
    k_loc = tuple(numpy.asarray(k_loc).tolist())
    assert len(k_loc) == len(distribution), "incorrect size of exponents"
    assert all([isinstance(k, int) for k in k_loc]), (
        "exponents must be integers: %s found" % type(k_loc[0]))

    if (distribution, order) not in MOMENTS_QUADS:
        MOMENTS_QUADS[distribution, order] = chaospy.generate_quadrature(
            order, distribution, rule=rule, **kwargs)
    X, W = MOMENTS_QUADS[distribution, order]

    if k_loc in distribution._mom_cache:
        return distribution._mom_cache[k_loc]

    out = float(numpy.sum(numpy.prod(X.T**k_loc, 1)*W))
    distribution._mom_cache[k_loc] = out
    return out


def approximate_density(
        distribution,
        idx,
        xloc,
        cache=None,
        tolerance=1e-7
):
    """
    Approximate the probability density function.

    Args:
        distribution (Distribution):
            Distribution in question. May not be an advanced variable.
        idx (int):
            The dimension to take approximation along.
        xloc (numpy.ndarray):
            Location coordinates. Requires that xloc.shape=(len(distribution), K).
        cache (Optional[Dict[Distribution, Tuple[numpy.ndarray, numpy.ndarray]]]):
            Current state in the evaluation graph. If omitted, assume that
            evaluations should be done from scratch.
        tolerance (float):
            Acceptable error level for the approximations

    Returns:
        numpy.ndarray: Local probability density function with
            ``out.shape == xloc.shape``. To calculate actual density function,
            evaluate ``numpy.prod(out, 0)``.

    Example:
        >>> distribution = chaospy.Normal(1000, 10)
        >>> xloc = numpy.array([990, 1000, 1010])
        >>> approximate_density(distribution, 0, xloc).round(4)
        array([0.0242, 0.0399, 0.0242])
        >>> distribution.pdf(xloc).round(4)
        array([0.0242, 0.0399, 0.0242])

    """
    cache = cache or {}
    xloc = numpy.asfarray(xloc)
    assert xloc.ndim == 1
    lower, upper = numpy.min(xloc), numpy.max(xloc)
    middle = .5*(lower+upper)
    tolerance = (numpy.where(xloc < middle, tolerance, -tolerance)*
                 numpy.clip(numpy.abs(xloc), 1, None))

    floc1 = distribution._get_fwd(xloc, idx, cache=cache.copy())
    floc2 = distribution._get_fwd(xloc+tolerance, idx, cache=cache.copy())
    floc = numpy.abs((floc2-floc1)/tolerance)

    assert floc.shape == xloc.shape
    return floc



"""Student-T distribution."""
import numpy
from scipy import special
import chaospy

from ..baseclass import SimpleDistribution, ShiftScaleDistribution


class student_t(SimpleDistribution):
    """Student-T distribution."""

    def __init__(self, a=1):
        super(student_t, self).__init__(dict(a=a))

    def _pdf(self, x, a):
        return special.gamma(.5*a+.5)*(1+x*x/a)**(-.5*a-.5) /\
                (numpy.sqrt(a*numpy.pi)*special.gamma(.5*a))

    def _cdf(self, x, a):
        return special.stdtr(a, x)

    def _ppf(self, q, a):
        return special.stdtrit(a, numpy.clip(q, 1e-12, 1-1e-12))

    def _mom(self, k, a):
        if numpy.any(a < k):
            raise ValueError("too high mom for student-t")
        out = special.gamma(.5*k+.5)* \
                special.gamma(.5*a-.5*k)*a**(.5*k)
        return numpy.where(k%2==0, out/(numpy.pi**.5*special.gamma(.5*a)), 0)

    def _ttr(self, k, a):
        return 0., k*a*(a-k+1.)/ ((a-2*k)*(a-2*k+2))

    def _lower(self, a):
        return special.stdtrit(a, 1e-12)

    def _upper(self, a):
        return special.stdtrit(a, 1-1e-12)


class StudentT(ShiftScaleDistribution):
    """
    (Non-central) Student-t distribution.

    Args:
        df (float, Distribution):
            Degrees of freedom.
        loc (float, Distribution):
            Location parameter.
        scale (float, Distribution):
            Scale parameter.

    Examples:
        >>> distribution = chaospy.StudentT(10)
        >>> distribution
        StudentT(10)
        >>> uloc = numpy.linspace(0, 1, 6)
        >>> uloc
        array([0. , 0.2, 0.4, 0.6, 0.8, 1. ])
        >>> xloc = distribution.inv(uloc)
        >>> xloc.round(3)
        array([-40.532,  -0.879,  -0.26 ,   0.26 ,   0.879,  40.532])
        >>> numpy.allclose(distribution.fwd(xloc), uloc)
        True
        >>> distribution.pdf(xloc).round(3)
        array([0.   , 0.258, 0.375, 0.375, 0.258, 0.   ])
        >>> distribution.sample(4).round(3)
        array([ 0.407, -1.278,  1.816, -0.046])
        >>> distribution.mom(1).round(3)
        0.0
        >>> distribution.ttr([0, 1, 2, 3]).round(3)
        array([[ 0.  ,  0.  ,  0.  ,  0.  ],
               [ 0.  ,  1.25,  3.75, 10.  ]])

    """

    def __init__(self, df=1, mu=0, sigma=1):
        super(StudentT, self).__init__(
            dist=student_t(df),
            scale=sigma,
            shift=mu,
        )
        self._repr_args = [df]
        self._repr_args += chaospy.format_repr_kwargs(mu=(mu, 0))
        self._repr_args += chaospy.format_repr_kwargs(sigma=(sigma, 1))



"""Bradford distribution."""
import numpy

from ..baseclass import SimpleDistribution, LowerUpperDistribution


class bradford(SimpleDistribution):
    """Standard Bradford distribution."""

    def __init__(self, c=1):
        super(bradford, self).__init__(dict(c=c))

    def _pdf(self, x, c):
        return  c/(c*x+1.)/numpy.log(1.+c)

    def _cdf(self, x, c):
        return numpy.log(1.+c*x)/numpy.log(c+1.)

    def _ppf(self, q, c):
        return ((1.+c)**q-1)/c

    def _lower(self, c):
        return 0.

    def _upper(self, c):
        return 1.


class Bradford(LowerUpperDistribution):
    """
    Bradford distribution.

    Args:
        shape (float, Distribution):
            Shape parameter
        lower (float, Distribution):
            Location of lower threshold
        upper (float, Distribution):
            Location of upper threshold

    Examples:
        >>> distribution = chaospy.Bradford(0.8, 4, 6)
        >>> distribution
        Bradford(0.8, lower=4, upper=6)
        >>> uloc = numpy.linspace(0, 1, 6)
        >>> uloc
        array([0. , 0.2, 0.4, 0.6, 0.8, 1. ])
        >>> xloc = distribution.inv(uloc)
        >>> xloc.round(3)
        array([4.   , 4.312, 4.663, 5.057, 5.501, 6.   ])
        >>> numpy.allclose(distribution.fwd(xloc), uloc)
        True
        >>> distribution.pdf(xloc).round(3)
        array([0.681, 0.605, 0.538, 0.478, 0.425, 0.378])
        >>> distribution.sample(4).round(3)
        array([5.171, 4.175, 5.87 , 4.819])

    """
    def __init__(self, shape=1, lower=0, upper=1):
        super(Bradford, self).__init__(
            dist=bradford(shape),
            lower=lower,
            upper=upper,
            repr_args=[shape],
        )



"""
Each sampling scheme can be accessed through the ``sample`` method on each
distribution. But in addition, they can also be created on the unit hyper-cube
using direct sampling functions. The frontend for all these functions is the
:func:`~chaospy.distributions.sampler.generator.generate_samples` function. It
allows for the same functionality as the ``sample`` method, but also support
some extra functionality by not being associated with a specific distribution.
For example::

    >>> samples = generate_samples(order=4)
    >>> samples.round(4)
    array([[0.6536, 0.115 , 0.9503, 0.4822]])

Custom domain::

    >>> samples = generate_samples(order=4, domain=[-1, 1])
    >>> samples.round(4)
    array([[ 0.7449, -0.5753, -0.9186, -0.2056]])
    >>> samples = generate_samples(order=4, domain=chaospy.Normal(0, 1))
    >>> samples.round(4)
    array([[-0.7286,  1.0016, -0.8166,  0.651 ]])

Use a custom sampling scheme::

    >>> generate_samples(order=4, rule="halton").round(4)
    array([[0.75 , 0.125, 0.625, 0.375]])

Multivariate case::

    >>> samples = generate_samples(order=4, domain=[[-1, 0], [0, 1]])
    >>> samples.round(4)
    array([[-0.6078, -0.8177, -0.2565, -0.9304],
           [ 0.8853,  0.9526,  0.9311,  0.4154]])
    >>> distribution = chaospy.J(chaospy.Normal(0, 1), chaospy.Uniform(0, 1))
    >>> samples = generate_samples(order=4, domain=distribution)
    >>> samples.round(4)
    array([[-1.896 ,  2.0975, -0.4135,  0.5437],
           [ 0.3619,  0.0351,  0.8551,  0.6573]])

Antithetic variates::

    >>> samples = generate_samples(order=8, rule="halton", antithetic=True)
    >>> samples.round(4)
    array([[0.75 , 0.25 , 0.125, 0.875, 0.625, 0.375, 0.375, 0.625]])

Multivariate antithetic variates::

    >>> samples = generate_samples(
    ...     order=8, domain=2, rule="hammersley", antithetic=True)
    >>> samples.round(4)
    array([[0.75 , 0.25 , 0.75 , 0.25 , 0.125, 0.875, 0.125, 0.875],
           [0.25 , 0.25 , 0.75 , 0.75 , 0.5  , 0.5  , 0.5  , 0.5  ]])

Here as with the ``sample`` method, the flag ``rule`` is used to determine
sampling scheme. The default ``rule="random"`` uses classical psuedo-random
samples created using ``numpy.random``.
"""
import logging
import numpy
from . import sequences, latin_hypercube

SAMPLER_NAMES = {
    "c": "chebyshev", "chebyshev": "chebyshev",
    "nc": "nested_chebyshev", "nested_chebyshev": "nested_chebyshev",
    "k": "korobov", "korobov": "korobov",
    "g": "grid", "grid": "grid",
    "ng": "nested_grid", "nested_grid": "nested_grid",
    "s": "sobol", "sobol": "sobol",
    "h": "halton", "halton": "halton",
    "m": "hammersley", "hammersley": "hammersley",
    "l": "latin_hypercube", "latin_hypercube": "latin_hypercube",
    "r": "random", "random": "random",
}
SAMPLER_FUNCTIONS = {
    "chebyshev": sequences.create_chebyshev_samples,
    "nested_chebyshev": sequences.create_nested_chebyshev_samples,
    "korobov": sequences.create_korobov_samples,
    "grid": sequences.create_grid_samples,
    "nested_grid": sequences.create_nested_grid_samples,
    "sobol": sequences.create_sobol_samples,
    "halton": sequences.create_halton_samples,
    "hammersley": sequences.create_hammersley_samples,
    "latin_hypercube": latin_hypercube.create_latin_hypercube_samples,
    "random": lambda order, dim: numpy.random.random((dim, order)),
}


def generate_samples(order, domain=1, rule="random", antithetic=None):
    """
    Sample generator.

    Args:
        order (int):
            Sample order. Determines the number of samples to create.
        domain (Distribution, int, numpy.ndarray):
            Defines the space where the samples are generated. If integer is
            provided, the space ``[0, 1]^domain`` will be used. If array-like
            object is provided, a hypercube it defines will be used. If
            distribution, the domain it spans will be used.
        rule (str):
            rule for generating samples. The various rules are listed in
            :mod:`chaospy.distributions.sampler.generator`.
        antithetic (tuple):
            Sequence of boolean values. Represents the axes to mirror using
            antithetic variable.
    """
    logger = logging.getLogger(__name__)

    if isinstance(domain, int):
        dim = domain
        trans = lambda x_data: x_data

    elif isinstance(domain, (tuple, list, numpy.ndarray)):
        domain = numpy.asfarray(domain)
        if len(domain.shape) < 2:
            dim = 1
        else:
            dim = len(domain[0])
        trans = lambda x_data: ((domain[1]-domain[0])*x_data.T + domain[0]).T

    else:
        dist = domain
        dim = len(dist)
        trans = dist.inv

    if antithetic is not None:

        from .antithetic import create_antithetic_variates
        antithetic = numpy.array(antithetic, dtype=bool).flatten()
        if antithetic.size == 1 and dim > 1:
            antithetic = numpy.repeat(antithetic, dim)

        size = numpy.sum(1*numpy.array(antithetic))
        order_saved = order
        order = int(numpy.log(order - dim))
        order = order if order > 1 else 1
        while order**dim < order_saved:
            order += 1
        trans_ = trans
        trans = lambda x_data: trans_(
            create_antithetic_variates(x_data, antithetic)[:, :order_saved])

    rule = SAMPLER_NAMES[rule.lower()]
    logger.debug("generating random samples using %s rule", rule)
    sampler = SAMPLER_FUNCTIONS[rule]
    x_data = trans(sampler(order=order, dim=dim))

    logger.debug("order: %d, dim: %d -> shape: %s", order, dim, x_data.shape)
    return x_data



"""Kernel density estimation baseclass."""
from __future__ import division

import numpy
import chaospy

from ..baseclass import Distribution


class KernelDensityBaseclass(Distribution):
    """Kernel density estimation baseclass."""

    def __init__(
            self,
            samples,
            h_mat=None,
            estimator_rule="scott",
            weights=None,
            rotation=None,
    ):
        """
        Args:
            samples (numpy.ndarray):
                The samples to generate density estimation. Assumed to have
                shape either as `(n_samples,)`, or `(n_dim, n_samples)`.
            h_mat (Optional[numpy.ndarray]):
                The H-matrix, also known as the smoothing matrix or bandwidth
                matrix. In one dimension it correspond to the square of the
                bandwidth parameters often used in the one-dimensional case.
                Assumes shape either something compatible with
                `(n_dim, n_dim)`, or `(n_dim, n_dim, n_samples)` in the case
                where each sample have their own H-matrix. If omitted, it is
                automatically calculated using `estimator_rule`.
            estimator_rule (str):
                Which method to use to select smoothing matrix from, assuming
                it is omitted. Choose from 'scott' and 'silverman'.
            weights (Optional[numpy.ndarray]):
                Weights of the samples. This must have the shape
                `(n_samples,)`. If omitted, each sample is assumed to be
                equally weighted.

        """
        samples = numpy.atleast_2d(samples)
        assert samples.ndim == 2

        # the scale is taken from Scott-92.
        # The Scott factor is taken from scipy docs.
        if h_mat is None:

            if estimator_rule in ("scott", "silverman"):
                qrange = numpy.quantile(samples, [0.25, 0.75], axis=1).ptp(axis=0)
                scale = numpy.min([numpy.std(samples, axis=1), qrange/1.34], axis=0)
                factor = samples.shape[1]
                if estimator_rule == "silverman":
                    factor *= (len(samples)+2)/4.
                factor **= -1./(len(samples)+4)
                covariance = numpy.diag(scale*factor)**2

            else:
                raise ValueError("unknown estimator rule: %s" % estimator_rule)

        else:
            covariance = numpy.asfarray(h_mat)
            if covariance.ndim in (0, 1):
                covariance = covariance*numpy.eye(len(samples))
        if covariance.ndim == 2:
            covariance = covariance[numpy.newaxis]
        else:
            covariance = numpy.rollaxis(covariance, 2, 0)
        assert covariance.shape[1:] == (len(samples), len(samples))

        if weights is None:
            weights = 1./samples.shape[1]
        self.weights = weights

        dependencies, _, rotation = chaospy.declare_dependencies(
            self, dict(), rotation=rotation,
            dependency_type="accumulate", length=len(samples),
        )

        self.samples = samples
        self.h_mat = covariance
        self._permute = numpy.eye(len(rotation), dtype=int)[rotation]
        self._pcovariance = numpy.matmul(numpy.matmul(
            self._permute, covariance), self._permute.T)
        cholesky = numpy.linalg.cholesky(self._pcovariance)
        self._fwd_transform = numpy.linalg.inv(cholesky)
        self._inv_transform = cholesky

        super(KernelDensityBaseclass, self).__init__(
            parameters={},
            dependencies=dependencies,
            rotation=rotation,
        )
        self._zloc = None
        self._kernel0 = None
        self._kernel1 = None

    def get_parameters(self, idx, cache, assert_numerical=True):
        parameters = super(KernelDensityBaseclass, self).get_parameters(
            idx, cache, assert_numerical=assert_numerical)
        if idx is None:
            del parameters["idx"]
        else:
            parameters["dim"] = dim = self._rotation.index(idx)
        return parameters

    def _pdf(self, x_loc, idx, dim, cache):
        """Kernel density function."""
        s, t = numpy.mgrid[:x_loc.shape[-1], :self.samples.shape[-1]]
        if not dim:
            samples = self.samples[idx, t]
            z_loc = ((x_loc[s]-samples)*self._fwd_transform[:, 0, 0])
            self._zloc = z_loc[:, :, numpy.newaxis]
            kernel = self._kernel(self._zloc)/self._inv_transform[:, 0, 0]
            self._kernel0 = self._kernel1 = kernel
            out = numpy.sum(kernel*self.weights, axis=-1)

        else:
            if self._zloc.shape[2] == dim+1:
                self._kernel0 = self._kernel1
            x_loc = [self._get_cache(dim_, cache, get=0)
                     for dim_ in self._rotation[:dim]] + [x_loc]
            x_loc = numpy.dstack([x[s] for x in x_loc])
            samples = numpy.dstack([self.samples[dim_, t]
                                    for dim_ in self._rotation[:dim+1]])
            zloc = numpy.sum((x_loc-samples)*self._fwd_transform[:, dim, :dim+1], -1)
            self._zloc = numpy.dstack([self._zloc[:, :, :dim], zloc])

            kernel = self._kernel(self._zloc)
            kernel *= (numpy.linalg.det(self._inv_transform[:, :dim, :dim])/
                       numpy.linalg.det(self._inv_transform[:, :dim+1, :dim+1]))
            out = (numpy.sum(kernel*self.weights, axis=-1)/
                   numpy.sum(self._kernel0*self.weights, axis=-1))
            self._kernel1, self._kernel0 = self._kernel0, kernel

        return out

    def _cdf(self, x_loc, idx, dim, cache):
        """Forward mapping."""
        s, t = numpy.mgrid[:x_loc.shape[-1], :self.samples.shape[-1]]
        if not dim:
            z_loc = (x_loc[s]-self.samples[idx, t])*self._fwd_transform[:, 0, 0]
            self._zloc = z_loc[:, :, numpy.newaxis]
            out = numpy.sum(self._ikernel(z_loc)*self.weights, axis=-1)
            assert out.shape == x_loc.shape, (out.shape, x_loc.shape)

        else:
            x_loc = [self._get_cache(dim_, cache, get=0)
                     for dim_ in self._rotation[:dim]] + [x_loc]
            x_loc = numpy.dstack([x[s] for x in x_loc])

            samples = numpy.dstack([self.samples[dim_, t]
                                    for dim_ in self._rotation[:dim+1]])
            zloc = numpy.sum((x_loc-samples)*self._fwd_transform[:, dim, :dim+1], -1)
            self._zloc = numpy.dstack([self._zloc[:, :, :dim], zloc])
            ikernel = self._kernel(self._zloc[:, :, :-1])*self._ikernel(self._zloc[:, :, -1])
            out = (numpy.sum(ikernel*self.weights, axis=-1)/
                   numpy.sum(self._kernel(self._zloc[:, :, :-1])*self.weights, axis=-1))
        return out

    def _ppf(self, u_loc, idx, dim, cache):
        """Inverse mapping."""
        xloc0 = numpy.quantile(self.samples[idx], u_loc)
        out = chaospy.approximate_inverse(
            self, idx, u_loc, xloc0=xloc0, cache=cache, iterations=1000)
        return out



"""Log-laplace distribution."""
import numpy
from scipy import special, misc

from ..baseclass import SimpleDistribution, ShiftScaleDistribution


class log_laplace(SimpleDistribution):
    """Log-laplace distribution."""

    def __init__(self, c):
        super(log_laplace, self).__init__(dict(c=c))

    def _pdf(self, x, c):
        cd2 = c/2.0
        c = numpy.where(x < 1, c, -c)
        return cd2*x**(c-1)

    def _cdf(self, x, c):
        return numpy.where(x < 1, 0.5*x**c, 1-0.5*x**(-c))

    def _ppf(self, q, c):
        return numpy.where(q < 0.5, (2.*q)**(1./c), (2*(1.-q))**(-1./c))

    def _lower(self, c):
        return 0.

    def _upper(self, c):
        return 2e12**(1./c)


class LogLaplace(ShiftScaleDistribution):
    """
    Log-laplace distribution

    Args:
        shape (float, Distribution):
            Shape parameter
        scale (float, Distribution):
            Scaling parameter
        shift (float, Distribution):
            Location parameter

    Examples:
        >>> distribution = chaospy.LogLaplace(5)
        >>> distribution
        LogLaplace(5)
        >>> uloc = numpy.linspace(0, 1, 6)
        >>> uloc
        array([0. , 0.2, 0.4, 0.6, 0.8, 1. ])
        >>> xloc = distribution.inv(uloc)
        >>> xloc.round(3)
        array([  0.   ,   0.833,   0.956,   1.046,   1.201, 288.54 ])
        >>> numpy.allclose(distribution.fwd(xloc), uloc)
        True
        >>> distribution.pdf(xloc).round(3)
        array([0.   , 1.201, 2.091, 1.913, 0.833, 0.   ])
        >>> distribution.sample(4).round(3)
        array([1.076, 0.745, 1.587, 0.993])

    """

    def __init__(self, shape=1, scale=1, shift=0):
        super(LogLaplace, self).__init__(
            dist=log_laplace(shape),
            scale=scale,
            shift=shift,
            repr_args=[shape],
        )



"""Generalized logistic type 1 distribution."""
import numpy
from scipy import special, misc

from ..baseclass import SimpleDistribution, ShiftScaleDistribution

class logistic(SimpleDistribution):
    """Generalized logistic type 1 distribution."""

    def __init__(self, c=1):
        super(logistic, self).__init__(dict(c=c))

    def _pdf(self, x, c):
        return numpy.e**-x/(1+numpy.e**-x)**(c+1)

    def _cdf(self, x, c):
        return (1+numpy.e**-x)**-c

    def _ppf(self, q, c):
        return -numpy.log(q**(-1./c)-1)

    def _lower(self, c):
        return -numpy.log(1e-12**(-1./c)-1)

    def _upper(self, c):
        return -numpy.log((1-1e-12)**(-1./c)-1)


class Logistic(ShiftScaleDistribution):
    """
    Generalized logistic type 1 distribution
    Sech squared distribution

    Args:
        skew (float, Distribution):
            Shape parameter
        shift (float, Distribution):
            Location parameter
        scale (float, Distribution):
            Scale parameter

    Examples:
        >>> distribution = chaospy.Logistic(15)
        >>> distribution
        Logistic(15)
        >>> uloc = numpy.linspace(0, 1, 6)
        >>> uloc
        array([0. , 0.2, 0.4, 0.6, 0.8, 1. ])
        >>> xloc = distribution.inv(uloc)
        >>> xloc.round(3)
        array([-1.67 ,  2.178,  2.765,  3.363,  4.201, 30.34 ])
        >>> numpy.allclose(distribution.fwd(xloc), uloc)
        True
        >>> distribution.pdf(xloc).round(3)
        array([0.   , 0.02 , 0.024, 0.02 , 0.012, 0.   ])
        >>> distribution.sample(4).round(3)
        array([3.549, 1.864, 5.682, 2.999])

    """
    def __init__(self, skew=1, shift=0, scale=1):
        super(Logistic, self).__init__(
            dist=logistic(skew),
            scale=scale,
            shift=shift,
            repr_args=[skew],
        )



"""Exponential Weibull distribution."""
import numpy

from ..baseclass import SimpleDistribution, ShiftScaleDistribution


class exponential_weibull(SimpleDistribution):
    """Exponential Weibull distribution."""

    def __init__(self, a=1, c=1):
        super(exponential_weibull, self).__init__(dict(a=a, c=c))

    def _pdf(self, x, a, c):
        exc = numpy.exp(-x**c)
        return a*c*(1-exc)**(a-1)*exc*x**(c-1)

    def _cdf(self, x, a, c):
        exm1c = -numpy.expm1(-x**c)
        return (exm1c)**a

    def _ppf(self, q, a, c):
        return (-numpy.log1p(-q**(1./a)))**(1./c)

    def _lower(self, a, c):
        return 0.

    def _upper(self, a, c):
        return (-numpy.log1p(-(1-1e-15)**(1./a)))**(1./c)


class ExponentialWeibull(ShiftScaleDistribution):
    """
    Exponential Weibull distribution.

    Args:
        alpha (float, Distribution):
            First shape parameter
        kappa (float, Distribution):
            Second shape parameter
        scale (float, Distribution):
            Scaling parameter
        shift (float, Distribution):
            Location parameter

    Examples:
        >>> distribution = chaospy.ExponentialWeibull(alpha=2, kappa=3)
        >>> distribution
        ExponentialWeibull(2, 3)
        >>> uloc = numpy.linspace(0, 1, 6)
        >>> uloc
        array([0. , 0.2, 0.4, 0.6, 0.8, 1. ])
        >>> xloc = distribution.inv(uloc)
        >>> xloc.round(3)
        array([0.   , 0.84 , 1.   , 1.142, 1.31 , 3.275])
        >>> numpy.allclose(distribution.fwd(xloc), uloc)
        True
        >>> distribution.pdf(xloc).round(3)
        array([0.   , 1.047, 1.396, 1.367, 0.972, 0.   ])
        >>> distribution.sample(4).round(3)
        array([1.182, 0.745, 1.544, 1.058])

    """
    def __init__(self, alpha=1, kappa=1, scale=1, shift=0):
        super(ExponentialWeibull, self).__init__(
            dist=exponential_weibull(alpha, kappa),
            scale=scale,
            shift=shift,
            repr_args=[alpha, kappa],
        )



r"""
A cumulative distribution function of an independent multivariate random
variable can be made dependent through a copula as follows:

.. math::
    F_{Q_0,\dots,Q_{D-1}} (q_0,\dots,q_{D-1}) =
    C(F_{Q_0}(q_0), \dots, F_{Q_{D-1}}(q_{D-1}))

where :math:`C` is the copula function, and :math:`F_{Q_i}` are marginal
distribution functions.  One of the more popular classes of copulas is the
Archimedean copulas.
.. \cite{sklar_random_1996}.
They are defined as follows:

.. math::
    C(u_1,\dots,u_n) =
    \phi^{[-1]} (\phi(u_1)+\dots+\phi(u_n)),

where :math:`\phi` is a generator and :math:`\phi^{[-1]}` is its
pseudo-inverse. Support for Archimedean copulas in `chaospy` is possible
through reformulation of the Rosenblatt transformation.  In two dimension, this
reformulation is as follows:

.. math::

    F_{U_0}(u_0) = \frac{C(u_0,1)}{C(1,1)}

    F_{U_1\mid U_0}(u_1\mid u_0) =
    \frac{\tfrac{\partial}{\partial u_0}
    C(u_0,u_1)}{\tfrac{\partial}{\partial u_0} C(u_0,1)}

This definition can also be generalized in to multiple variables using the
formula provided by Nelsen 1999.
.. cite:: nelsen_introduction_1999

The definition of the Rosenblatt transform can require multiple
differentiations.  An analytical formulation is usually not feasible, so the
expressions are estimated using difference scheme similar to the one outlined
for probability density function defined in :ref:`distributions`. The accurate
might therefore be affected.

Since copulas are meant as a replacement for Rosenblatt
transformation, it is usually assumed that the distribution it is
used on is stochastically independent.
However in the definition of a copula does not actually require it, and sine
the Rosenblatt transformation allows for it, multiple copulas can be stacked
together in `chaospy`.
"""
import numpy
import chaospy

from .distribution import Distribution


class CopulaDistribution(Distribution):
    """Baseclass for all Copulas."""

    def __init__(self, dist, trans, rotation=None, repr_args=None):
        r"""
        Args:
            dist (Distribution):
                Distribution to wrap the copula around.
            trans (Distribution):
                The copula wrapper `[0,1]^D \into [0,1]^D`.

        """
        assert len(dist) == len(trans), "Copula length missmatch"
        accumulant = set()
        dependencies = [deps.copy() for deps in dist._dependencies]
        for idx, _ in sorted(enumerate(trans._dependencies), key=lambda x: len(x[1])):
            accumulant.update(dist._dependencies[idx])
            dependencies[idx] = accumulant.copy()

        super(CopulaDistribution, self).__init__(
            parameters=dict(dist=dist, trans=trans),
            dependencies=dependencies,
            rotation=rotation,
            repr_args=repr_args,
        )

    def get_parameters(self, idx, cache, assert_numerical=True):
        parameters = super(CopulaDistribution, self).get_parameters(
            idx, cache, assert_numerical=assert_numerical)
        if idx is None:
            del parameters["idx"]
        return parameters

    def _lower(self, idx, dist, trans, cache):
        return dist._get_lower(idx, cache=cache)

    def _upper(self, idx, dist, trans, cache):
        return dist._get_upper(idx, cache=cache)

    def _cdf(self, xloc, idx, dist, trans, cache):
        output = dist._get_fwd(xloc, idx, cache=cache)
        output = trans._get_fwd(output, idx, cache=cache)
        return output

    def _ppf(self, qloc, idx, dist, trans, cache):
        qloc = trans._get_inv(qloc, idx, cache=cache)
        xloc = dist._get_inv(qloc, idx, cache=cache)
        return xloc

    def _pdf(self, xloc, idx, dist, trans, cache):
        density = dist._get_pdf(xloc, idx, cache=cache.copy())
        return trans._get_pdf(
            dist._get_fwd(xloc, idx, cache=cache), idx, cache=cache)*density

    def _mom(self, xloc, dist, trans, cache):
        raise chaospy.UnsupportedFeature("Copula not supported.")

    def _ttr(self, xloc, idx, dist, trans, cache):
        raise chaospy.UnsupportedFeature("Copula not supported.")



"""
Generate the quadrature abscissas and weights for simple grid.

Available to ensure that discrete distributions works along side
continuous ones.

Example usage
-------------

The first few orders with linear growth rule::

    >>> distribution = chaospy.DiscreteUniform(-2, 2)
    >>> for order in [0, 1, 2, 3, 4, 5, 9]:
    ...     abscissas, weights = chaospy.generate_quadrature(
    ...         order, distribution, rule="discrete")
    ...     print(order, abscissas.round(3), weights.round(3))
    0 [[0]] [1.]
    1 [[-1  1]] [0.5 0.5]
    2 [[-2  0  2]] [0.333 0.333 0.333]
    3 [[-2 -1  1  2]] [0.25 0.25 0.25 0.25]
    4 [[-2 -1  0  1  2]] [0.2 0.2 0.2 0.2 0.2]
    5 [[-2 -1  0  1  2]] [0.2 0.2 0.2 0.2 0.2]
    9 [[-2 -1  0  1  2]] [0.2 0.2 0.2 0.2 0.2]

As the accuracy of discrete distribution plateau when all contained values are
included, there is no reason to increase the number of nodes after this point.

The first few orders with exponential growth rule where the nodes are nested::

    >>> distribution = chaospy.DiscreteUniform(0, 10)
    >>> for order in [0, 1, 2, 3, 4]:
    ...     abscissas, weights = chaospy.generate_quadrature(
    ...         order, distribution, rule="discrete", growth=True)
    ...     print(order, abscissas)
    0 [[5]]
    1 [[1 5 9]]
    2 [[1 3 5 7 9]]
    3 [[ 0  1  3  4  5  6  7  9 10]]
    4 [[ 0  1  2  3  4  5  6  7  8  9 10]]
"""
import numpy
import chaospy

from .combine import combine_quadrature
from .grid import quad_grid


def quad_discrete(order, domain=(0, 1), growth=False, segments=1):
    """
    Generate quadrature abscissas and weights for discrete distributions.

    Same as regular grid, but `order` plateau at the `upper-lower-1`.
    At this order, finite state discrete distributions are analytically
    correct, and higher order will make the accuracy worsen.

    Args:
        order (int, numpy.ndarray):
            Quadrature order.
        domain (chaospy.distributions.baseclass.Distribution, numpy.ndarray):
            Either distribution or bounding of interval to integrate over.
        growth (bool):
            if true sets the growth rule for the quadrature rule to only
            include orders that enhances nested samples.

    Returns:
        (numpy.ndarray, numpy.ndarray):
            The quadrature points and weights. The points are
            equi-spaced grid on the interior of the domain bounds.
            The weights are all equal to `1/len(weights[0])`.
            Either distribution or bounding of interval to integrate over.

    Examples:
        >>> distribution = chaospy.DiscreteUniform(-2, 2)
        >>> abscissas, weights = chaospy.quad_discrete(4, distribution)
        >>> abscissas.round(4)
        array([[-2., -1.,  0.,  1.,  2.]])
        >>> weights.round(4)
        array([0.2, 0.2, 0.2, 0.2, 0.2])
        >>> abscissas, weights = chaospy.quad_discrete(9, distribution)
        >>> abscissas.round(4)
        array([[-2., -1.,  0.,  1.,  2.]])
        >>> weights.round(4)
        array([0.2, 0.2, 0.2, 0.2, 0.2])

    """
    if isinstance(domain, chaospy.Distribution):
        abscissas, weights = quad_discrete(order, (domain.lower, domain.upper), growth=growth, segments=segments)
        weights *= domain.pdf(abscissas).flatten()
        weights /= numpy.sum(weights)
        return abscissas, weights

    del segments  # Basically segments doesn't do much here
    if growth:
        order = numpy.where(order > 0, 2**(order), 0)

    order = numpy.atleast_1d(order)
    order, lower, upper = numpy.broadcast_arrays(order, domain[0], domain[1])
    assert order.ndim == 1, "too many dimensions"

    order_max = numpy.round(upper-lower).astype(int)-1
    order = numpy.where(order > order_max, order_max, order)

    return quad_grid(order, (lower, upper))



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['chaospy',
 'chaospy.chol',
 'chaospy.descriptives',
 'chaospy.descriptives.correlation',
 'chaospy.descriptives.sensitivity',
 'chaospy.distributions',
 'chaospy.distributions.baseclass',
 'chaospy.distributions.collection',
 'chaospy.distributions.copulas',
 'chaospy.distributions.kernel',
 'chaospy.distributions.operators',
 'chaospy.distributions.sampler',
 'chaospy.distributions.sampler.sequences',
 'chaospy.external',
 'chaospy.orthogonal',
 'chaospy.quadrature',
 'chaospy.quadrature.genz_keister',
 'chaospy.quadrature.recurrence']

package_data = \
{'': ['*']}

install_requires = \
['numpoly==1.0.6', 'numpy', 'scipy']

extras_require = \
{':python_version >= "2.7" and python_version < "3.0"': ['functools32']}

setup_kwargs = {
    'name': 'chaospy',
    'version': '4.0b2',
    'description': 'Numerical tool for perfroming uncertainty quantification',
    'long_description': '.. image:: https://github.com/jonathf/chaospy/raw/master/docs/.static/chaospy_logo.svg\n   :height: 200 px\n   :width: 200 px\n   :align: center\n\n|circleci| |codecov| |readthedocs| |pypi|\n\n|license| |python-versions| |downloads|\n\n.. |circleci| image:: https://img.shields.io/circleci/build/github/jonathf/chaospy/master\n    :target: https://circleci.com/gh/jonathf/chaospy/tree/master\n.. |codecov| image:: https://img.shields.io/codecov/c/github/jonathf/chaospy\n    :target: https://codecov.io/gh/jonathf/chaospy\n.. |readthedocs| image:: https://img.shields.io/readthedocs/chaospy\n    :target: http://chaospy.readthedocs.io/en/master/?badge=master\n.. |pypi| image:: https://img.shields.io/pypi/v/chaospy\n    :target: https://pypi.org/project/chaospy\n.. |license| image:: https://img.shields.io/pypi/l/chaospy\n    :target: https://opensource.org/licenses/MIT\n.. |python-versions| image:: https://img.shields.io/pypi/pyversions/chaospy\n.. |downloads| image:: https://img.shields.io/pypi/dw/chaospy\n    :target: https://pypistats.org/packages/chaospy\n\nChaospy is a numerical tool for performing uncertainty quantification using\npolynomial chaos expansions and advanced Monte Carlo methods implemented in\nPython.\n\n.. contents:: Table of Contents:\n\nInstallation\n------------\n\nInstallation should be straight forward::\n\n    pip install chaospy\n\nAnd you should be ready to go.\n\nExample Usage\n-------------\n\n``chaospy`` is created to be simple and modular. A simple script to implement\npoint collocation method will look as follows:\n\n.. code-block:: python\n\n    import numpy\n    import chaospy\n\nWrap your code in a function:\n\n.. code-block:: python\n\n    coordinates = numpy.linspace(0, 10, 100)\n    def foo(coordinates, params):\n        """Function to do uncertainty quantification on."""\n        param_init, param_rate = params\n        return param_init*numpy.e**(-param_rate*coordinates)\n\nConstruct a multivariate probability distribution:\n\n.. code-block:: python\n\n    distribution = chaospy.J(chaospy.Uniform(1, 2), chaospy.Uniform(0.1, 0.2))\n\nConstruct polynomial chaos expansion:\n\n.. code-block:: python\n\n    polynomial_expansion = chaospy.generate_expansion(8, distribution)\n\nGenerate random samples from for example Halton low-discrepancy sequence:\n\n.. code-block:: python\n\n    samples = distribution.sample(1000, rule="halton")\n\nEvaluate function for each sample:\n\n.. code-block:: python\n\n    evals = numpy.array([foo(coordinates, sample) for sample in samples.T])\n\nBring the parts together using point collocation method:\n\n.. code-block:: python\n\n    foo_approx = chaospy.fit_regression(\n        polynomial_expansion, samples, evals)\n\nDerive statistics from model approximation:\n\n.. code-block:: python\n\n    expected = chaospy.E(foo_approx, distribution)\n    deviation = chaospy.Std(foo_approx, distribution)\n    sobol_main = chaospy.Sens_m(foo_approx, distribution)\n    sobol_total = chaospy.Sens_t(foo_approx, distribution)\n\nFor a more extensive guides on what is going on, see the `tutorial collection`_.\n\n.. _tutorial collection: https://chaospy.readthedocs.io/en/master/tutorials\n\nRelated Projects\n----------------\n\nChaospy is being used in other related projects that requires uncertainty\nquantification components ``chaospy`` provides.\n\n+-----------------+-----------------------------------------------------------+\n| `easyVVUQ`_     | Library designed to facilitate verification, validation   |\n|                 | and uncertainty quantification.                           |\n+-----------------+-----------------------------------------------------------+\n| `STARFiSh`_     | Shell-based, scientific simulation program                |\n|                 | for blood flow in mammals.                                |\n+-----------------+-----------------------------------------------------------+\n| `Profit`_       | Probabilistic response model fitting via interactive      |\n|                 | tools.                                                    |\n+-----------------+-----------------------------------------------------------+\n| `UncertainPy`_  | Uncertainty quantification and sensitivity analysis,      |\n|                 | tailored towards computational neuroscience.              |\n+-----------------+-----------------------------------------------------------+\n| `SparseSpACE`_  | Spatially adaptive combination technique targeted to      |\n|                 | solve high dimensional numerical integration.             |\n+-----------------+-----------------------------------------------------------+\n\n.. _easyVVUQ: https://github.com/UCL-CCS/EasyVVUQ\n.. _STARFiSh: https://www.ntnu.no/starfish\n.. _Profit: https://github.com/redmod-team/profit\n.. _UncertainPy: https://github.com/simetenn/uncertainpy\n.. _SparseSpACE: https://github.com/obersteiner/sparseSpACE\n\nAlso a few shout-outs:\n\n+--------------+--------------------------------------------------------------+\n| `OpenTURNS`_ | Thanks to `Régis Lebrun`_ for both proposing a collaboration |\n|              | and creating an initial implementation of both               |\n|              | `Chaospy Compatibility`_ in `OpenTURNS`_ and                 |\n|              | `OpenTURNS Compatibility`_ in ``chaospy``.                   |\n+--------------+--------------------------------------------------------------+\n| `orthopy`_   | Thanks to `Nico Schlömer`_ for providing the implementation  |\n| `quadpy`_    | for several of the quadrature integration methods.           |\n+--------------+--------------------------------------------------------------+\n| ``UQRF``     | Thanks to `Florian Künzner`_ for providing the               |\n|              | implementation for `sample distribution`_.                   |\n+--------------+--------------------------------------------------------------+\n\n.. _OpenTURNS: http://openturns.github.io/openturns/latest\n.. _Régis Lebrun: https://github.com/regislebrun\n.. _Chaospy Compatibility: http://openturns.github.io/openturns/latest/user_manual/_generated/openturns.ChaospyDistribution.html\n.. _OpenTURNS Compatibility: https://chaospy.readthedocs.io/en/master/recipes/external.html#module-chaospy.external.openturns_\n.. _orthopy: https://github.com/nschloe/orthopy\n.. _quadpy: https://github.com/nschloe/quadpy\n.. _Nico Schlömer: https://github.com/nschloe\n.. _Florian Künzner: https://github.com/flo2k\n.. _sample distribution: https://chaospy.readthedocs.io/en/master/recipes/external.html#module-chaospy.external.samples\n\nQuestions and Contributions\n---------------------------\n\nPlease feel free to\n`file an issue <https://github.com/jonathf/chaospy/issues>`_ for:\n\n* bug reporting\n* asking questions related to usage\n* requesting new features\n* wanting to contribute with code\n\nIf you are using this software in work that will be published, please cite the\njournal article: `Chaospy: An open source tool for designing methods of\nuncertainty quantification <http://dx.doi.org/10.1016/j.jocs.2015.08.008>`_\n',
    'author': 'Jonathan Feinberg',
    'author_email': None,
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/jonathf/chaospy',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
}


setup(**setup_kwargs)



"""Pearson's correlation matrix."""
import numpy
from scipy.stats import spearmanr
import numpoly
import chaospy


def Corr(poly, dist=None, **kws):
    """
    Correlation matrix of a distribution or polynomial.

    Args:
        poly (numpoly.ndpoly, Distribution):
            Input to take correlation on. Must have ``len(poly)>=2``.
        dist (Distribution):
            Defines the space the correlation is taken on.  It is ignored if
            ``poly`` is a distribution.

    Returns:
        (numpy.ndarray):
            Correlation matrix with
            ``correlation.shape == poly.shape+poly.shape``.

    Examples:
        >>> distribution = chaospy.MvNormal(
        ...     [3, 4], [[2, .5], [.5, 1]])
        >>> chaospy.Corr(distribution).round(4)
        array([[1.    , 0.3536],
               [0.3536, 1.    ]])
        >>> q0 = chaospy.variable()
        >>> poly = chaospy.polynomial([q0, q0**2])
        >>> distribution = chaospy.Normal()
        >>> chaospy.Corr(poly, distribution).round(4)
        array([[1., 0.],
               [0., 1.]])

    """
    if isinstance(poly, chaospy.Distribution):
        poly, dist = numpoly.variable(len(poly)), poly
    else:
        poly = numpoly.polynomial(poly)

    if not poly.shape:
        return numpy.ones((1, 1))

    cov = chaospy.Cov(poly, dist, **kws)
    var = numpy.diag(cov)
    vvar = numpy.sqrt(numpy.outer(var, var))
    return numpy.where(vvar > 0, cov/vvar, 0)



"""
Gaussian Quadratures Rules
--------------------------

:ref:`gaussian`
    The classical Gaussian quadrature scheme applied on any probability
    distribution.
:ref:`gauss_legendre`
    Same as :ref:`gaussian` for uniform distribution, but applicable to other
    distribution by incorporating the probability density as part of the
    function to be integrated.
:ref:`gauss_patterson`
    Extension of Gauss-Legendre rule. Valid to order 8.
:ref:`gauss_kronrod`
    Extension to the Gauss-Patterson rule to include most distribution and any
    order.
:ref:`gauss_lobatto`
    Gaussian quadrature rule that enforces the endpoints to be included in the
    rule.
:ref:`gauss_radau`
    Gaussian quadrature rule that enforces that a single fixed point to be
    included in the rule.

Non-Gaussian Quadrature Rules
-----------------------------

:ref:`clenshaw_curtis`
    Chebyshev nodes with endpoints included.
:ref:`fejer`
    Chebyshev nodes without endpoints included.
:ref:`leja`
    Fully nested quadrature method.
:ref:`genz_keister`
    Genz-Keizter 16 rule. Nested. Valid to order 8.
:ref:`newton_cotes`
    Numerical integration rule based on fixed width abscissas.
"""
import numpy
import chaospy

from .combine import combine

from .clenshaw_curtis import quad_clenshaw_curtis
from .discrete import quad_discrete
from .fejer import quad_fejer
from .gaussian import quad_gaussian
from .gauss_patterson import quad_gauss_patterson
from .gauss_legendre import quad_gauss_legendre
from .gauss_lobatto import quad_gauss_lobatto
from .gauss_kronrod import quad_gauss_kronrod
from .gauss_radau import quad_gauss_radau
from .genz_keister import quad_genz_keister
from .grid import quad_grid
from .leja import quad_leja
from .newton_cotes import quad_newton_cotes

QUAD_NAMES = {
    "c": "clenshaw_curtis", "clenshaw_curtis": "clenshaw_curtis",
    "f": "fejer", "fejer": "fejer",
    "g": "gaussian", "gaussian": "gaussian",
    "e": "gauss_legendre", "gauss_legendre": "gauss_legendre",
    "l": "gauss_lobatto", "gauss_lobatto": "gauss_lobatto",
    "k": "gauss_kronrod", "gauss_kronrod": "gauss_kronrod",
    "p": "gauss_patterson", "gauss_patterson": "gauss_patterson",
    "r": "gauss_radau", "gauss_radau": "gauss_radau",
    "z": "genz_keister", "genz_keister": "genz_keister",
    "j": "leja", "leja": "leja",
    "n": "newton_cotes", "newton_cotes": "newton_cotes",
    "d": "discrete", "discrete": "discrete",
    "i": "grid", "grid": "grid",
}
QUAD_FUNCTIONS = {
    "clenshaw_curtis": quad_clenshaw_curtis,
    "fejer": quad_fejer,
    "gaussian": quad_gaussian,
    "gauss_kronrod": quad_gauss_kronrod,
    "gauss_legendre": quad_gauss_legendre,
    "gauss_lobatto": quad_gauss_lobatto,
    "gauss_patterson": quad_gauss_patterson,
    "gauss_radau": quad_gauss_radau,
    "genz_keister": quad_genz_keister,
    "leja": quad_leja,
    "newton_cotes": quad_newton_cotes,
    "discrete": quad_discrete,
    "grid": quad_grid,
}


def generate_quadrature(
        order,
        dist,
        rule="",
        sparse=False,
        accuracy=100,
        growth=None,
        segments=1,
        recurrence_algorithm="",
):
    """
    Numerical quadrature node and weight generator.

    Args:
        order (int):
            The order of the quadrature.
        dist (chaospy.distributions.baseclass.Distribution):
            The distribution which density will be used as weight function.
        rule (str, Sequence[str]):
            Rule for generating abscissas and weights. If one name is provided,
            that rule is applied to all dimensions. If multiple names, each
            rule is positionally applied to each dimension. If omitted,
            ``clenshaw_curtis`` is applied to all continuous dimensions, and
            ``discrete`` to all discrete ones.
        sparse (bool):
            If True used Smolyak's sparse grid instead of normal tensor product
            grid.
        accuracy (int):
            If gaussian is set, but the dist provided in domain does not
            provide an analytical TTR, ac sets the approximation order for the
            descitized Stieltje's method.
        growth (bool, None):
            If True sets the growth rule for the quadrature rule to only
            include orders that enhances nested samples. Defaults to the same
            value as ``sparse`` if omitted.
        segments (int):
            Split intervals into N subintervals and create a patched
            quadrature based on the segmented quadrature. Can not be lower than
            `order`. If 0 is provided, default to square root of `order`.
            Nested samples only exist when the number of segments are fixed.
        recurrence_algorithm (str):
            Name of the algorithm used to generate abscissas and weights in
            case of Gaussian quadrature scheme. If omitted, ``analytical`` will
            be tried first, and ``stieltjes`` used if that fails.

    Returns:
        (numpy.ndarray, numpy.ndarray):
            Abscissas and weights created from full tensor grid rule. Flatten
            such that ``abscissas.shape == (len(dist), len(weights))``.

    Examples:
        >>> distribution = chaospy.Iid(chaospy.Normal(0, 1), 2)
        >>> abscissas, weights = generate_quadrature(
        ...     1, distribution, rule=("gaussian", "fejer"))
        >>> abscissas.round(3)
        array([[-1.  , -1.  ,  1.  ,  1.  ],
               [-4.11,  4.11, -4.11,  4.11]])
        >>> weights.round(3)
        array([0.25, 0.25, 0.25, 0.25])

    """
    if not rule:
        if isinstance(dist, chaospy.J):
            rule = [
                ("discrete" if dist_.interpret_as_integer else "clenshaw_curtis")
                for dist_ in dist
            ]
        else:
            rule = ("discrete" if dist.interpret_as_integer
                    else "clenshaw_curtis")
    if sparse:
        from . import sparse_grid
        return sparse_grid.construct_sparse_grid(
            order, dist, rule=rule, accuracy=accuracy, growth=growth)

    if len(dist) == 1 or dist.stochastic_dependent:
        assert isinstance(rule, str), "dependencies require rule consistency"
        abscissas, weights = _generate_quadrature(
            order=order,
            dist=dist,
            rule=rule,
            growth=growth,
            segments=segments,
            accuracy=accuracy,
            recurrence_algorithm=recurrence_algorithm,
        )

    else:
        if isinstance(rule, str):
            rule = [rule]*len(dist)
        assert len(rule) == len(dist), (
            "rules and distribution length does not match.")
        assert all(isinstance(rule_, str) for rule_ in rule)

        order = numpy.ones(len(dist), dtype=int)*order
        abscissas, weights = zip(*[
            _generate_quadrature(
                order=order_,
                dist=dist_,
                rule=rule_,
                growth=growth,
                segments=segments,
                accuracy=accuracy,
                recurrence_algorithm=recurrence_algorithm,
            )
            for order_, dist_, rule_ in zip(order, dist, rule)
        ])
        abscissas = combine([abscissa.T for abscissa in abscissas]).T
        weights = numpy.prod(combine([weight.T for weight in weights]), -1)

    assert abscissas.shape == (len(dist), len(weights))
    if dist.interpret_as_integer:
        abscissas = numpy.round(abscissas).astype(int)
    return abscissas, weights


def _generate_quadrature(order, dist, rule, **kwargs):

    if isinstance(dist, chaospy.OperatorDistribution):

        args = ("left", "right")
        right_dist = isinstance(dist._parameters["right"], chaospy.Distribution)
        args = args if right_dist else args[::-1]
        assert not isinstance(dist._parameters[args[0]], chaospy.Distribution)
        const = numpy.asarray(dist._parameters[args[0]])

        if isinstance(dist, chaospy.Add):
            dist = dist._parameters[args[1]]
            abscissas, weights = _generate_quadrature(
                order=order, dist=dist, rule=rule, **kwargs)
            abscissas = (abscissas.T+const.T).T
            return abscissas, weights

        elif isinstance(dist, chaospy.Multiply):
            dist = dist._parameters[args[1]]
            abscissas, weights = _generate_quadrature(
                order=order, dist=dist, rule=rule, **kwargs)
            abscissas = (abscissas.T*const.T).T
            return abscissas, weights

    rule = QUAD_NAMES[rule.lower()]
    parameters = {}

    if rule in ("clenshaw_curtis", "fejer", "newton_cotes", "discrete"):
        parameters.update(growth=kwargs["growth"], segments=kwargs["segments"])

    if rule in ("gaussian", "gauss_kronrod", "gauss_radau", "gauss_lobatto"):
        parameters.update(accuracy=kwargs["accuracy"],
                          recurrence_algorithm=kwargs["recurrence_algorithm"])

    quad_function = QUAD_FUNCTIONS[rule]
    abscissas, weights = quad_function(order, dist, **parameters)
    return abscissas, weights



"""Percentile function."""
import numpy


def Perc(poly, q, dist, sample=10000, **kws):
    """
    Percentile function.

    Note that this function is an empirical function that operates using Monte
    Carlo sampling.

    Args:
        poly (numpoly.ndpoly):
            Polynomial of interest.
        q (numpy.ndarray):
            positions where percentiles are taken. Must be a number or an
            array, where all values are on the interval ``[0, 100]``.
        dist (Distribution):
            Defines the space where percentile is taken.
        sample (int):
            Number of samples used in estimation.

    Returns:
        (numpy.ndarray):
            Percentiles of ``poly`` with ``Q.shape=poly.shape+q.shape``.

    Examples:
        >>> dist = chaospy.J(chaospy.Gamma(1, 1), chaospy.Normal(0, 2))
        >>> q0, q1 = chaospy.variable(2)
        >>> poly = chaospy.polynomial([0.05*q0, 0.2*q1, 0.01*q0*q1])
        >>> chaospy.Perc(poly, [0, 5, 50, 95, 100], dist).round(2)
        array([[ 0.  , -3.29, -4.54],
               [ 0.  , -0.64, -0.04],
               [ 0.03, -0.01, -0.  ],
               [ 0.15,  0.66,  0.04],
               [ 1.38,  3.29,  4.54]])

    """
    shape = poly.shape
    poly = poly.flatten()

    q = numpy.array(q)/100.
    dim = len(dist)

    # Interior
    Z = dist.sample(sample, **kws)
    if dim==1:
        Z = (Z,)
        q = numpy.array([q])
    poly1 = poly(*Z)

    # Min/max
    ext = numpy.mgrid[(slice(0, 2, 1), )*dim].reshape(dim, 2**dim).T
    ext = numpy.where(ext, dist.lower, dist.upper).T
    poly2 = poly(*ext)
    poly2 = numpy.array([_ for _ in poly2.T if not numpy.any(numpy.isnan(_))]).T

    # Finish
    if poly2.shape:
        poly1 = numpy.concatenate([poly1, poly2], -1)
    samples = poly1.shape[-1]
    poly1.sort()
    out = poly1.T[numpy.asarray(q*(samples-1), dtype=int)]
    out = out.reshape(q.shape + shape)

    return out



"""Conditional expected value."""
from itertools import product

import numpy
import numpoly

from . import expected


def E_cond(poly, freeze, dist, **kws):
    """
    Conditional expected value operator.

    1st order statistics of a polynomial on a given probability space
    conditioned on some of the variables.

    Args:
        poly (numpoly.ndpoly):
            Polynomial to find conditional expected value on.
        freeze (numpy.ndarray):
            Boolean values defining the conditional variables. True values
            implies that the value is conditioned on, e.g. frozen during the
            expected value calculation.
        dist (Distribution) :
            The distributions of the input used in ``poly``.

    Returns:
        (numpoly.ndpoly) :
            Same as ``poly``, but with the variables not tagged in ``frozen``
            integrated away.

    Examples:
        >>> q0, q1 = chaospy.variable(2)
        >>> poly = chaospy.polynomial([1, q0, q1, 10*q0*q1-1])
        >>> poly
        polynomial([1, q0, q1, 10*q0*q1-1])
        >>> dist = chaospy.J(chaospy.Gamma(1, 1), chaospy.Normal(0, 2))
        >>> chaospy.E_cond(poly, [1, 0], dist)
        polynomial([1.0, q0, 0.0, -1.0])
        >>> chaospy.E_cond(poly, [0, 1], dist)
        polynomial([1.0, 1.0, q1, 10.0*q1-1.0])
        >>> chaospy.E_cond(poly, [1, 1], dist)
        polynomial([1, q0, q1, 10*q0*q1-1])
        >>> chaospy.E_cond(poly, [0, 0], dist)
        polynomial([1.0, 1.0, 0.0, -1.0])

    """
    poly = numpoly.set_dimensions(poly, len(dist))
    if not poly.isconstant:
        return poly.tonumpy()
    assert not dist.stochastic_dependent, dist

    freeze = numpoly.polynomial(freeze)
    if freeze.isconstant():
        freeze = freeze.tonumpy().astype(bool)
    else:
        poly, freeze = numpoly.align_exponents(poly, freeze)
        freeze = numpy.isin(poly.keys, freeze.keys)

    # decompose into frozen and unfrozen part
    poly = numpoly.decompose(poly)
    unfrozen = poly(**{
        ("q%d" % idx): 1 for idx, keep in enumerate(freeze) if keep})
    frozen = poly(**{
        ("q%d" % idx): 1 for idx, keep in enumerate(freeze) if not keep})

    # if no unfrozen, poly will return numpy.ndarray instead of numpoly.ndpoly
    if not isinstance(unfrozen, numpoly.ndpoly):
        return numpoly.sum(frozen, 0)

    # Remove frozen coefficients, such that poly == sum(frozen*unfrozen) holds
    for key in unfrozen.keys:
        unfrozen[key] = unfrozen[key] != 0

    return numpoly.sum(frozen*expected.E(unfrozen, dist), 0)



"""
Truncation.

Example usage
-------------

Simple distribution to start with::

    >>> distribution = chaospy.Normal(0, 1)
    >>> distribution.inv([0.9, 0.99, 0.999]).round(4)
    array([1.2816, 2.3263, 3.0902])

Same distribution, but with a right-side truncation::

    >>> right_trunc = chaospy.Trunc(chaospy.Normal(0, 1), 1)
    >>> right_trunc
    Trunc(Normal(mu=0, sigma=1), 1)
    >>> right_trunc.inv([0.9, 0.99, 0.999]).round(4)
    array([0.6974, 0.9658, 0.9965])

Same, but with left-side truncation::

    >>> left_trunc = chaospy.Trunc(1, chaospy.Normal(0, 1))
    >>> left_trunc
    Trunc(1, Normal(mu=0, sigma=1))
    >>> left_trunc.inv([0.001, 0.01, 0.1]).round(4)
    array([1.0007, 1.0066, 1.0679])

"""
import numpy
import chaospy

from ..baseclass import Distribution, OperatorDistribution


class Trunc(OperatorDistribution):
    """Truncation."""

    def __init__(self, left, right):
        """
        Constructor.

        Args:
            left (Distribution, numpy.ndarray):
                Left hand side.
            right (Distribution, numpy.ndarray):
                Right hand side.
        """
        repr_args = [left, right]
        exclusion = None
        if isinstance(left, Distribution):
            if left.stochastic_dependent:
                raise chaospy.StochasticallyDependentError(
                    "Joint distribution with dependencies not supported.")
        else:
            left = numpy.atleast_1d(left)
        if isinstance(right, Distribution):
            if right.stochastic_dependent:
                raise chaospy.StochasticallyDependentError(
                    "Joint distribution with dependencies not supported.")
        else:
            right = numpy.atleast_1d(right)
        super(Trunc, self).__init__(
            left=left,
            right=right,
            repr_args=repr_args,
        )

    def _lower(self, idx, left, right, cache):
        """
        Distribution lower bound.

        Examples:
            >>> chaospy.Trunc(chaospy.Uniform(), 0.6).lower
            array([0.])
            >>> chaospy.Trunc(0.6, chaospy.Uniform()).lower
            array([0.6])
        """
        del right
        if isinstance(left, Distribution):
            left = left._get_lower(idx, cache=cache)
        return left

    def _upper(self, idx, left, right, cache):
        """
        Distribution lower bound.

        Examples:
            >>> chaospy.Trunc(chaospy.Uniform(), 0.6).upper
            array([0.6])
            >>> chaospy.Trunc(0.6, chaospy.Uniform()).upper
            array([1.])
        """
        del left
        if isinstance(right, Distribution):
            right = right._get_upper(idx, cache=cache)
        return right

    def _cdf(self, xloc, idx, left, right, cache):
        """
        Cumulative distribution function.

        Example:
            >>> chaospy.Uniform().fwd([-0.5, 0.3, 0.7, 1.2])
            array([0. , 0.3, 0.7, 1. ])
            >>> chaospy.Trunc(chaospy.Uniform(), 0.4).fwd([-0.5, 0.2, 0.8, 1.2])
            array([0. , 0.5, 1. , 1. ])
            >>> chaospy.Trunc(0.6, chaospy.Uniform()).fwd([-0.5, 0.2, 0.8, 1.2])
            array([0. , 0. , 0.5, 1. ])
        """
        if isinstance(left, Distribution):
            left = left._get_cache(idx, cache, get=0)
        if isinstance(right, Distribution):
            right = right._get_cache(idx, cache, get=0)
        if isinstance(left, Distribution):
            right = (numpy.array(right).T*numpy.ones(xloc.shape).T).T
            uloc1 = left._get_fwd(right, idx, cache=cache.copy())
            uloc2 = left._get_fwd(xloc, idx, cache=cache)
            out = uloc2/uloc1
        else:
            left = (numpy.array(left).T*numpy.ones(xloc.shape).T).T
            uloc1 = right._get_fwd(left, idx, cache=cache.copy())
            uloc2 = right._get_fwd(xloc, idx, cache=cache)
            out = (uloc2-uloc1)/(1-uloc1)
        return out

    def _pdf(self, xloc, idx, left, right, cache):
        """
        Probability density function.

        Example:
            >>> dist = chaospy.Trunc(chaospy.Uniform(), 0.6)
            >>> dist.pdf([-0.25, 0.25, 0.5, 0.75, 1.25])
            array([0.        , 1.66666667, 1.66666667, 0.        , 0.        ])
            >>> dist = chaospy.Trunc(chaospy.Uniform(), 0.4)
            >>> dist.pdf([-0.25, 0.25, 0.5, 0.75, 1.25])
            array([0. , 2.5, 0. , 0. , 0. ])
            >>> dist = chaospy.Trunc(0.4, chaospy.Uniform())
            >>> dist.pdf([-0.25, 0.25, 0.5, 0.75, 1.25])
            array([0.        , 0.        , 1.66666667, 1.66666667, 0.        ])
            >>> dist = chaospy.Trunc(0.6, chaospy.Uniform())
            >>> dist.pdf([-0.25, 0.25, 0.5, 0.75, 1.25])
            array([0. , 0. , 0. , 2.5, 0. ])
        """
        if isinstance(left, Distribution):
            left = left._get_cache(idx, cache, get=0)
        if isinstance(right, Distribution):
            right = right._get_cache(idx, cache, get=0)
        if isinstance(left, Distribution):
            right = (numpy.array(right).T*numpy.ones(xloc.shape).T).T
            uloc1 = left._get_fwd(right, idx,  cache=cache.copy())
            uloc2 = left._get_pdf(xloc, idx, cache=cache)
            out = uloc2/uloc1
        else:
            left = (numpy.array(left).T*numpy.ones(xloc.shape).T).T
            uloc1 = right._get_fwd(left, idx, cache=cache.copy())
            uloc2 = right._get_pdf(xloc, idx, cache=cache)
            out = uloc2/(1-uloc1)
        return out

    def _ppf(self, q, idx, left, right, cache):
        """
        Point percentile function.

        Example:
            >>> chaospy.Uniform().inv([0.1, 0.2, 0.9])
            array([0.1, 0.2, 0.9])
            >>> chaospy.Trunc(chaospy.Uniform(), 0.4).inv([0.1, 0.2, 0.9])
            array([0.04, 0.08, 0.36])
            >>> chaospy.Trunc(0.6, chaospy.Uniform()).inv([0.1, 0.2, 0.9])
            array([0.64, 0.68, 0.96])
        """
        if isinstance(left, Distribution):
            left = left._get_cache(idx, cache, get=0)
        if isinstance(right, Distribution):
            right = right._get_cache(idx, cache, get=0)
        if isinstance(left, Distribution):
            right = (numpy.array(right).T*numpy.ones(q.shape).T).T
            uloc = left._get_fwd(right, idx, cache=cache.copy())
            out = left._get_inv(q*uloc, idx, cache=cache)
        else:
            left = (numpy.array(left).T*numpy.ones(q.shape).T).T
            uloc = right._get_fwd(left, idx, cache=cache.copy())
            out = right._get_inv(q*(1-uloc)+uloc, idx, cache=cache)
        return out



"""
Historically, the syntax for ``chaospy``'s distributions is inspired by the
``scipy.stats`` distributions using the syntax. Over time though, they have
developed into different entities making them incompatible. The
:class:`ScipyStatsDist` exists to restore this incompatibility again.

Basic usage of the :class:`ScipyStatsDist` constructor involves just passing
the ``scipy.stats`` distributions as input argument::

    >>> from scipy.stats import norm
    >>> st_distribution = norm(0, 1)
    >>> distribution = chaospy.ScipyStatsDist(st_distribution)
    >>> distribution
    ScipyStatsDist(scipy.stats.norm(0, 1))

This distribution then behaves as a normal ``chaospy`` distribution::

    >>> distribution.pdf([-1, 0, 1]).round(4)
    array([0.242 , 0.3989, 0.242 ])
    >>> distribution.mom([0, 1, 2])
    array([1., 0., 1.])

Currently the wrapper is limited to only support univariate distributions.
"""
import numpy
from scipy.stats._distn_infrastructure import rv_frozen

from ..distributions import SimpleDistribution


class ScipyStatsDist(SimpleDistribution):
    """
    One dimensional ``scipy.stats`` distribution.

    Args:
        distribution (openturns.Distribution):
            1D distribution created with ``scipy.stats``.
    """

    def __init__(self, distribution):
        assert isinstance(distribution, rv_frozen), (
            "Expected frozen distribution from ``scipy.stats``.")
        assert not hasattr(distribution, "dim") or distribution.dim == 1, (
            "Only one-dimensional ``scipy.stats`` models supported.")
        name = distribution.dist.__class__.__name__.replace("_gen", "")
        args = ", ".join([str(arg) for arg in distribution.args])
        super(ScipyStatsDist, self).__init__(
            repr_args=["scipy.stats.%s(%s)" % (name, args)],
        )
        self.distribution = distribution

    def _pdf(self, x_loc):
        return self.distribution.pdf(x_loc)

    def _cdf(self, x_loc):
        return self.distribution.cdf(x_loc)

    def _ppf(self, q_loc):
        return self.distribution.ppf(q_loc)

    def _lower(self):
        return self.distribution.interval(1-1e-14)[0]

    def _upper(self):
        return self.distribution.interval(1-1e-14)[1]

    def _mom(self, k_loc):
        return self.distribution.moment(int(k_loc))



"""Double gamma distribution."""
import numpy
from scipy import special

from ..baseclass import SimpleDistribution, ShiftScaleDistribution


class double_gamma(SimpleDistribution):
    """Double gamma distribution."""

    def __init__(self, a):
        super(double_gamma, self).__init__(dict(a=a))

    def _pdf(self, x, a):
        ax = abs(x)
        return 1.0/(2*special.gamma(a))*ax**(a-1.0) * numpy.exp(-ax)

    def _cdf(self, x, a):
        fac = 0.5*special.gammainc(a,abs(x))
        return numpy.where(x>0,0.5+fac,0.5-fac)

    def _ppf(self, q, a):
        fac = special.gammainccinv(a, 1-abs(2*q-1))
        out = numpy.where(q > 0.5, fac, -fac)
        return out

    def _lower(self, a):
        return -special.gammainccinv(a, 2e-15)

    def _upper(self, a):
        return special.gammainccinv(a, 2e-15)


class DoubleGamma(ShiftScaleDistribution):
    """
    Double gamma distribution.

    Args:
        shape (float, Distribution):
            Shape parameter
        scale (float, Distribution):
            Scaling parameter
        shift (float, Distribution):
            Location parameter

    Examples:
        >>> distribution = chaospy.DoubleGamma(shape=1.5)
        >>> distribution
        DoubleGamma(1.5)
        >>> uloc = numpy.linspace(0, 1, 6)
        >>> uloc
        array([0. , 0.2, 0.4, 0.6, 0.8, 1. ])
        >>> xloc = distribution.inv(uloc)
        >>> xloc.round(3)
        array([-35.769,  -1.473,  -0.503,   0.503,   1.473,  35.769])
        >>> numpy.allclose(distribution.fwd(xloc), uloc)
        True
        >>> distribution.pdf(xloc).round(3)
        array([0.   , 0.157, 0.242, 0.242, 0.157, 0.   ])
        >>> distribution.sample(4).round(3)
        array([ 0.727, -2.154,  3.132, -0.138])

    """

    def __init__(self, shape=1, scale=1, shift=0):
        super(DoubleGamma, self).__init__(
            dist=double_gamma(shape),
            scale=scale,
            shift=shift,
            repr_args=[shape],
        )



"""Frontend function for generating polynomial expansions."""
from .three_terms_recurrence import orth_ttr
from .cholesky import orth_chol
from .gram_schmidt import orth_gs
from .lagrange import lagrange_polynomial

EXPANSION_NAMES = {
    "ttr": "three_terms_recurrence", "three_terms_recurrence": "three_terms_recurrence",
    "chol": "cholesky", "cholesky": "cholesky",
    "gs": "gram_schmidt", "gram_schmidt": "gram_schmidt",
}
EXPANSION_FUNCTIONS = {
    "three_terms_recurrence": orth_ttr,
    "cholesky": orth_chol,
    "gram_schmidt": orth_gs,
}


def generate_expansion(
        order,
        dist,
        rule="three_terms_recurrence",
        normed=False,
        graded=True,
        reverse=True,
        cross_truncation=1.,
        sort=None,
        **kws
):
    """
    Create orthogonal polynomial expansion.

    This function is a frontend wrapper for the three methods for creating
    orthogonal polynomials:

    +------------------------+-------------------------------------------------+
    | Algorithm              | Description                                     |
    +------------------------+-------------------------------------------------+
    | three_terms_recurrence | Three terms recurrence coefficients generated   |
    |                        | using Stieltjes and Golub-Welsch method. The    |
    |                        | most stable of the methods, but do not work on  |
    |                        | dependent distributions.                        |
    +------------------------+-------------------------------------------------+
    | gram_schmidt           | Gram-Schmidt orthogonalization method applied   |
    |                        | on polynomial expansions. Know for being        |
    |                        | numerically unstable.                           |
    +------------------------+-------------------------------------------------+
    | cholesky               | Orthogonalization through decorrelation of the  |
    |                        | covariance matrix. Uses Gill-King's Cholesky    |
    |                        | decomposition method for higher numerical       |
    |                        | stability. Still not scalable to high number of |
    |                        | dimensions.                                     |
    +------------------------+-------------------------------------------------+

    Args:
        order (int):
            Order of polynomial expansion.
        dist (Distribution):
            Distribution space where polynomials are orthogonal. If the method
            ``dist._ttr`` exists, it will be used.
        rule (str):
            The orthogonalization method used.
        normed (bool):
            If True orthonormal polynomials will be used.
        graded (bool):
            Graded sorting, meaning the indices are always sorted by the index
            sum. E.g. ``q0**2*q1**2*q2**2`` has an exponent sum of 6, and will
            therefore be consider larger than both ``q0**2*q1*q2``,
            ``q0*q1**2*q2`` and ``q0*q1*q2**2``,
            which all have exponent sum of 5.
        reverse (bool):
            Reverse lexicographical sorting meaning that ``q0*q1**3`` is
            considered bigger than ``q0**3*q1``, instead of the opposite.
        retall (bool):
            If true return numerical stabilized norms as well. Roughly the same
            as ``cp.E(orth**2, dist)``.
        cross_truncation (float):
            Use hyperbolic cross truncation scheme to reduce the number of
            terms in expansion. only include terms where the exponents ``K``
            satisfied the equation
            ``order >= sum(K**(1/cross_truncation))**cross_truncation``.

    Returns:
        (numpoly.ndpoly, numpy.ndarray):
            Orthogonal polynomial expansion. norms of the orthogonal
            expansion on the form ``E(orth**2, dist)``. Calculated using
            recurrence coefficients for stability.

    Examples:
        >>> distribution = chaospy.Normal()
        >>> expansion, norms = generate_expansion(
        ...     3, distribution, retall=True)
        >>> expansion
        polynomial([1.0, q0, q0**2-1.0, q0**3-3.0*q0])
        >>> norms
        array([1., 1., 2., 6.])

    """
    name = EXPANSION_NAMES[rule.lower()]
    expansion_function = EXPANSION_FUNCTIONS[name]
    return expansion_function(order, dist=dist, normed=normed, graded=graded,
                              reverse=reverse, sort=sort,
                              cross_truncation=cross_truncation, **kws)



"""Create an N-dimensional unit polynomial basis."""
import logging

import numpy
import numpoly


def basis(start, stop=None, dim=1, cross_truncation=1., sort="G"):
    """
    Create an N-dimensional unit polynomial basis.

    Args:
        start (int, numpy.ndarray):
            the minimum polynomial to include. If int is provided, set as
            lowest total order.  If array of int, set as lower order along each
            axis.
        stop (int, numpy.ndarray):
            the maximum shape included. If omitted:
            ``stop <- start; start <- 0`` If int is provided, set as largest
            total order. If array of int, set as largest order along each axis.
        dim (int):
            dim of the basis. Ignored if array is provided in either start or
            stop.
        cross_truncation (float):
            Use hyperbolic cross truncation scheme to reduce the number of
            terms in expansion.

    Returns:
        (numpoly.ndpoly) : Polynomial array.

    Examples:
        >>> chaospy.basis(2, dim=2)
        polynomial([1, q1, q0, q1**2, q0*q1, q0**2])
        >>> chaospy.basis(2, dim=4, cross_truncation=0)
        polynomial([1, q3, q2, q1, q0, q3**2, q2**2, q1**2, q0**2])
        >>> chaospy.basis(2, 2, dim=2, cross_truncation=numpy.inf)
        polynomial([q1**2, q0**2, q0*q1**2, q0**2*q1, q0**2*q1**2])

    """
    logger = logging.getLogger(__name__)
    logger.warning("chaospy.basis is deprecated; use chaospy.monomial instead")

    if stop is None:
        start, stop = 0, start
    dim = max(numpy.asarray(start).size, numpy.asarray(stop).size, dim)
    out = numpoly.monomial(
        start=start,
        stop=numpy.array(stop)+1,
        reverse="R" not in sort.upper(),
        graded="G" in sort.upper(),
        cross_truncation=cross_truncation,
        names=numpoly.symbols("q:%d" % dim, asarray=True),
    )
    if "I" in sort.upper():
        out = out[::-1]
    return out
